#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
Functionality for "lazy loading". A lazy loaded module is registered in sys.modules as
loaded, but is not actually run intil needed, i.e. when an attribute which requires the
module to be imported is accessed. This is useful for optimization purposes, but needs to
be used with caution.

Since the actual import of a lazy loaded module is postponend until it is actually needed,
exceptions occuring during the actual import can happen in a location where you may not
have anticipated it. Threfore, exceptions raised during the lazy import are wrapped in a
dedicated exception class, LazyImportError, to avoid potential import exceptions being
misclassified.
'''

import sys
from types import ModuleType
from typing import Set
from importlib._bootstrap import _ImportLockContext
import importlib
from nomad import config


_not_yet_imported_lazy_module_names: Set[str] = set()


class LazyImportError(Exception):
    pass


class _LazyModule(ModuleType):
    '''
    Base class for lazy modules.
    '''
    def __getattribute__(self, attr):
        '''
        Overrides the standard method, to trigger the actual import of the lazy module
        when a non-trivial attribute is accessed.

        Note: when we actually import the module, we will replace this method with the
        standard method, thereby reverting to standard behaviour.
        '''
        if attr not in ('__name__', '__class__', '__spec__', '__repr__', '__file__'):
            try:
                # In case the attribute we are trying to access is actually another
                # lazy-loaded module, just return it.
                name = '%s.%s' % (self.__name__, attr)
                return sys.modules[name]
            except KeyError:
                pass
            # No, we have to actually load the module now!
            _actually_import(self.__name__)
        return ModuleType.__getattribute__(self, attr)  # Standard functionality

    def __setattr__(self, attr, value):
        '''
        Overrides the standard method, to trigger the actual import of the lazy module
        when any attribute is set on the module.

        Note: when we actually import the module, we will replace this method with the
        standard method, thereby reverting to standard behaviour.
        '''
        _actually_import(self.__name__)
        return ModuleType.__setattr__(self, attr, value)  # Standard functionality


def _actually_import(module_name):
    '''
    Actually import the lazy module. Also make sure that all its parent modules are
    imported if needed - and they need to be imported in the right order.
    '''
    with _ImportLockContext():
        parts = module_name.split('.')
        base_name = ''
        for part in parts:
            if base_name:
                base_name += '.'
            base_name += part
            if base_name in _not_yet_imported_lazy_module_names:
                # This level is a lazy module, and it has not yet been loaded. Load it!
                _not_yet_imported_lazy_module_names.remove(base_name)
                module = sys.modules[base_name]
                # Restore __getattribute__ and __setattr__ to original functionality
                module_class = type(module)
                module_class.__getattribute__ = ModuleType.__getattribute__
                module_class.__setattr__ = ModuleType.__setattr__
                # Remove the fake __file__ attribute we set initially
                ModuleType.__delattr__(module, '__file__')
                # Actually import the module
                try:
                    importlib.reload(module)
                except Exception as e:
                    # Wrap the exception, to avoid potential exception misclassification.
                    err_msg = f'Error occured during loading of lazy module {base_name}: {e}'
                    raise LazyImportError(err_msg)


def _create_lazy_module(module_name):
    '''
    Create a dedicated class and instantiate it, and adds it to sys.modules
    '''
    class _LazyModuleSubclass(_LazyModule):
        def __repr__(self):
            return 'Lazily-loaded module %s' % self.__name__
    module = _LazyModuleSubclass(module_name)
    ModuleType.__setattr__(module, '__file__', None)
    sys.modules[module_name] = module
    _not_yet_imported_lazy_module_names.add(module_name)
    return module


def lazy_module(module_name):
    '''
    Call this to "lazily" import a module. Subsequent calls to import will succeed
    immediately, without the module actually being imported. The module is imported
    first when it is actually used (by accessing an attribute which requires the
    module to really be imported).

    The lazy import functionality can also be disabled using the setting
        nomad.config.enable_lazy_import = False
    When disabled, this method does nothing, and modules are imported "as usual".
    '''
    if not config.enable_lazy_import:
        return

    if module_name not in sys.modules:
        # Create a lazy module object and add it, without really loading it.
        # Also add a lazy module object for all parent modules, if needed.
        with _ImportLockContext():
            module = _create_lazy_module(module_name)
            while True:
                parent_module_name, _, submodule_name = module_name.rpartition('.')
                if not parent_module_name:
                    break
                # Fetch or create parent_module
                if parent_module_name in sys.modules:
                    parent_module = sys.modules[parent_module_name]
                    parent_was_already_created = True
                else:
                    parent_module = _create_lazy_module(parent_module_name)
                    parent_was_already_created = False
                # Set module as an attribute on the parent_module
                ModuleType.__setattr__(parent_module, submodule_name, module)
                if parent_was_already_created:
                    break
                # Parent had to be lazy loaded too -> we need to also check parent's parent.
                module = parent_module
                module_name = parent_module_name
    return



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from flask import Blueprint, request
import os.path

gui_folder = os.path.abspath(os.path.join(os.path.dirname(__file__), 'static/gui'))
configuired_gui_folder = os.path.join(gui_folder, '../.gui_configured')
if os.path.exists(configuired_gui_folder):
    gui_folder = configuired_gui_folder
blueprint = Blueprint('gui', __name__, static_url_path='/', static_folder=gui_folder)


@blueprint.after_request
def add_header(response):
    if request.url.endswith('index.html'):
        response.headers['Cache-Control'] = 'no-cache, no-store, must-revalidate'
    if request.url.endswith('.js'):
        response.headers['Cache-Control'] = 'no-cache, must-revalidate'
    return response



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import numpy as np
import logging
from datetime import datetime

from nomad.units import ureg
from nomad.parsing.parser import FairdiParser
from nomad.parsing.file_parser.text_parser import TextParser, Quantity
from nomad.datamodel.metainfo.common_dft import Run, Method, XCFunctionals, System,\
    SingleConfigurationCalculation, ScfIteration, Dos, SamplingMethod


class OutParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        re_float = r'[\d\.\-\+Ee]+'

        def str_to_labels_positions(val_in):
            val = np.transpose([v.split() for v in val_in.strip().split('\n')])
            return val[1], np.array(val[2:5].T, dtype=np.dtype(np.float64)) * ureg.bohr

        calculation_quantities = [
            Quantity(
                'labels_positions',
                r' Index Symbol\s*x\s*\(bohr\).+([\s\S]+?)\n *\n',
                convert=False, str_operation=str_to_labels_positions),
            Quantity(
                'lattice_vectors',
                r'Lattice vectors \(bohr\)\s*([\s\S]+?)\n *\n',
                convert=False,
                str_operation=lambda x: np.array(
                    [v.split()[1:4] for v in x.strip().split('\n')], dtype=np.dtype(np.float64))),
            Quantity(
                'total_charge',
                rf'Total System Charge\s*({re_float})', dtype=float, unit=ureg.elementary_charge),
            Quantity(
                'band_engine_input',
                r'Band Engine Input\s*\-+\s*([\s\S]+?)\n *\n',
                sub_parser=TextParser(quantities=[
                    Quantity('basis', r' Basis\s*Type\s*(.+)')])),
            Quantity(
                'model_parameters',
                r'M O D E L   P A R A M E T E R S\s*\=+([\s\S]+?)\={10}',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'dft_potential',
                        r'FUNCTIONAL POTENTIAL \(scf\)([\s\S]+?)DENSITY',
                        sub_parser=TextParser(quantities=[
                            Quantity('LDA', r'LDA\:\s*([\w ]+)', flatten=False),
                            Quantity('GGA', r'Gradient Corrections\:\s*([\w ]+)', flatten=False),
                            Quantity('MGGA', r'Meta-GGA\:\s*([\w ]+)', flatten=False)])),
                    Quantity(
                        'spin',
                        r'SPIN\s*\(restricted.+\s*.+\:\s*(\w+)')])),
            Quantity(
                'fermi_energy',
                rf'Fermi Energy\:\s*({re_float})\s*a\.u\.', dtype=float, unit=ureg.hartree),
            Quantity(
                'energies',
                r'E N E R G Y   A N A L Y S I S([\s\S]+?)\={90}',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'electronic_kinetic_energy',
                        rf'Kinetic\s*({re_float})', dtype=float, unit=ureg.hartree),
                    Quantity(
                        'energy_XC',
                        rf'XC\s*({re_float})', dtype=float, unit=ureg.hartree),
                    Quantity(
                        'energy_electrostatic',
                        rf'Electrostatic\s*({re_float})', dtype=float, unit=ureg.hartree),
                    Quantity(
                        'energy_total',
                        rf'Final bond energy \(.+\)\s*({re_float})', dtype=float, unit=ureg.hartree)])),
            Quantity(
                'energy_total',
                rf'Energy\s*\(hartree\)\s*({re_float})', dtype=float, unit=ureg.hartree),
            Quantity(
                'atom_forces',
                r'FINAL GRADIENTS([\s\S]+?)\n *\n',
                convert=False, str_operation=lambda x: np.array(
                    [v.split()[2:5] for v in x.strip().split('\n')],
                    dtype=np.dtype(np.float64)) * ureg.hartree / ureg.bohr),
            Quantity(
                'self_consistency',
                r'S C F   P R O C E D U R E\s*\*\s*\*+\s*([\s\S]+?Self consistent error.+)',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'energy_change',
                        rf'cyc\=\s*\d+\s*err\=\s*({re_float})',
                        repeats=True, dtype=float, unit=ureg.hartree)])),
            Quantity(
                'total_dos',
                r'TOTALDOS([\s\S]+?)ENDINPUT',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'dos',
                        rf'\n *({re_float}) *({re_float}) *({re_float})*',
                        dtype=np.dtype(np.float64), repeats=True)]))
            # TODO add code specific metainfo
        ]

        self._quantities = [
            Quantity('program_version', r'\*\s*r(\d+ \d{4}\-\d\d\-\d\d)', flatten=False),
            Quantity('time_start', r'RunTime\:\s*(\w{3}\d+\-\d{4}\s*\d+\:\d+\:\d+)', flatten=False),
            Quantity(
                'single_point',
                r'SINGLE POINT CALCULATION \*([\s\S]+?)(:?Timing|\Z)',
                sub_parser=TextParser(quantities=calculation_quantities)),
            Quantity(
                'geometry_optimization',
                r'GEOMETRY OPTIMIZATION\s*\*([\s\S]+?)(:?Timing|\Z)',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'iteration',
                        r'Geometry Convergence after Step\s*([\s\S]+?(?:dE\(predicted\)\:.+|\Z))',
                        repeats=True, sub_parser=TextParser(quantities=calculation_quantities)),
                    Quantity(
                        'geometry_optimization_method',
                        r'Optimization Method\s*(.+)', flatten=False),
                    Quantity(
                        'geometry_optimization_threshold_force',
                        rf'Maximum gradient\s*({re_float})',
                        dtype=float, unit=ureg.hartree / ureg.bohr),
                    Quantity(
                        'geometry_optimization_energy_change',
                        rf'Maximum energy change allowe\s*({re_float})',
                        dtype=float, unit=ureg.hartree),
                    Quantity(
                        'geometry_optimization_geometry_change',
                        rf'Maximum step allowed\s*({re_float})',
                        dtype=float, unit=ureg.bohr)]))
            # TODO add other calculation types
        ]


class BandParser(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/band', code_name='BAND',
            code_homepage='https://www.scm.com/product/band_periodicdft/',
            mainfile_contents_re=r' +\* +Amsterdam Density Functional +\(ADF\)')

        self.out_parser = OutParser()

    def init_parser(self):
        self.out_parser.mainfile = self.mainfile
        self.out_parser.logger = self.logger

    def parse_configurations(self):
        sec_run = self.archive.section_run[0]

        def parse_scc(source):
            sec_scc = sec_run.m_create(SingleConfigurationCalculation)

            for key in ['energy_total', 'atom_forces']:
                val = source.get(key)
                if val is not None:
                    setattr(sec_scc, key, val)

            # energy contributions
            for key, val in source.get('energies', {}).items():
                if val is not None:
                    setattr(sec_scc, key, val)

            # self consistency
            for energy_change in source.get('self_consistency', {}).get('energy_change', []):
                sec_scf = sec_scc.m_create(ScfIteration)
                sec_scf.energy_change_scf_iteration = energy_change

            # dos
            total_dos = source.get('total_dos', {}).get('dos')
            if total_dos is not None:
                total_dos = np.transpose(total_dos)
                sec_dos = sec_scc.m_create(Dos)
                sec_dos.dos_energies = total_dos[0] * ureg.hartree
                sec_dos.dos_values = (total_dos[1:] * (1 / ureg.hartree)).to('1/J').magnitude

            return sec_scc

        def parse_system(source):
            sec_system = sec_run.m_create(System)

            labels_positions = source.get('labels_positions')
            if labels_positions is not None:
                sec_system.atom_labels = labels_positions[0]
                sec_system.atom_positions = labels_positions[1]

            lattice_vectors = source.get('lattice_vectors')
            if lattice_vectors is not None:
                lattice_vectors = list(lattice_vectors)
                pbc = [True, True, True]
                for n in range(len(lattice_vectors), 3):
                    lattice_vectors.append([0, 0, 0])
                    pbc[n] = False
                sec_system.lattice_vectors = lattice_vectors * ureg.bohr
                sec_system.configuration_periodic_dimensions = pbc

            return sec_system

        def parse_method(source):
            sec_method = sec_run.m_create(Method)
            sec_method.electronic_structure_method = 'DFT'

            dft_potential = source.get('model_parameters', {}).get('dft_potential', {})
            # TODO provide mapping
            for xc_type in ['LDA', 'GGA', 'MGGA']:
                functionals = dft_potential.get(xc_type, '').split()
                kind = ['XC'] if len(functionals) == 1 else ['X', 'C']
                for n, functional in enumerate(functionals):
                    sec_functional = sec_method.m_create(XCFunctionals)
                    functional = functional.rstrip('x').rstrip('c').upper()
                    sec_functional.XC_functional_name = '%s_%s_%s' % (xc_type, kind[n], functional)

            spin = source.get('model_parameters', {}).get('spin')
            sec_method.number_of_spin_channels = 2 if spin == ('UNrestricted') else 1

            if source.get('total_charge') is not None:
                sec_method.total_charge = source.total_charge

            return sec_method

        def parse_calculation(source):
            if source is None:
                return

            sec_scc = parse_scc(source)
            sec_system = parse_system(source)
            sec_method = parse_method(source)
            sec_scc.single_configuration_calculation_to_system_ref = sec_system
            sec_scc.single_configuration_to_calculation_method_ref = sec_method

        parse_calculation(self.out_parser.get('single_point'))

        geometry_opt = self.out_parser.get('geometry_optimization')
        if geometry_opt is not None:
            sec_sampling_method = sec_run.m_create(SamplingMethod)
            for key, val in geometry_opt.items():
                setattr(sec_sampling_method, key, val)
            for iteration in geometry_opt.get('iteration', []):
                parse_calculation(iteration)

    def parse(self, filepath, archive, logger):
        self.mainfile = os.path.abspath(filepath)
        self.archive = archive
        self.logger = logger if logger is not None else logging.getLogger(__name__)

        self.init_parser()

        sec_run = self.archive.m_create(Run)
        sec_run.program_name = 'BAND'
        sec_run.program_basis_set_type = 'numeric AOs'

        if self.out_parser.get('program_version') is not None:
            sec_run.program_version = self.out_parser.program_version

        if self.out_parser.get('time_start') is not None:
            dt = datetime.strptime(self.out_parser.time_start, '%b%d-%Y %H:%M:%S') - datetime(1970, 1, 1)
            sec_run.time_run_date_start = dt.total_seconds()

        self.parse_configurations()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Optional, Union, Dict, Iterator, Any, List, Set, IO
from fastapi import APIRouter, Depends, Path, status, HTTPException, Request, Query as QueryParameter
from fastapi.responses import StreamingResponse
import os.path
import io
import json
import orjson
import magic
import gzip
import lzma

from nomad import search, files, config, utils
from nomad.utils import strip
from nomad.archive import RequiredReader, RequiredValidationError, ArchiveQueryError

from .auth import create_user_dependency
from ..utils import create_streamed_zipfile, File, create_responses
from ..models import (
    EntryPagination, WithQuery, MetadataRequired, EntriesMetadataResponse, EntriesMetadata,
    EntryMetadataResponse, query_parameters, metadata_required_parameters, Files, Query,
    entry_pagination_parameters, files_parameters, User, Owner, HTTPExceptionModel, EntriesRaw,
    EntriesRawResponse, EntriesRawDownload, EntryRaw, EntryRawFile, EntryRawResponse,
    EntriesArchiveDownload, EntryArchiveResponse, EntriesArchive, EntriesArchiveResponse,
    ArchiveRequired, EntryArchiveRequest)


router = APIRouter()
default_tag = 'entries'
metadata_tag = 'entries/metadata'
raw_tag = 'entries/raw'
archive_tag = 'entries/archive'

logger = utils.get_logger(__name__)


_bad_owner_response = status.HTTP_401_UNAUTHORIZED, {
    'model': HTTPExceptionModel,
    'description': strip('''
        Unauthorized. The given owner requires authorization,
        but no or bad authentication credentials are given.''')}

_bad_id_response = status.HTTP_404_NOT_FOUND, {
    'model': HTTPExceptionModel,
    'description': strip('''
        Entry not found. The given id does not match any entry.''')}

_bad_path_response = status.HTTP_404_NOT_FOUND, {
    'model': HTTPExceptionModel,
    'description': strip('File or directory not found.')}

_raw_download_response = 200, {
    'content': {'application/zip': {}},
    'description': strip('''
        A zip file with the requested raw files. The file is streamed.
        The content length is not known in advance.
    ''')}

_raw_download_file_response = 200, {
    'content': {'application/octet-stream': {}},
    'description': strip('''
        A byte stream with raw file contents. The content length is not known in advance.
        If the whole file is requested, the mime-type might be more specific, depending
        on the file contents.
    ''')}

_archive_download_response = 200, {
    'content': {'application/zip': {}},
    'description': strip('''
        A zip file with the requested archive files. The file is streamed.
        The content length is not known in advance.
    ''')}


_bad_archive_required_response = status.HTTP_400_BAD_REQUEST, {
    'model': HTTPExceptionModel,
    'description': strip('''
        The given required specification could not be understood.''')}


def perform_search(*args, **kwargs):
    try:
        return search.search(*args, **kwargs)
    except search.AuthenticationRequiredError as e:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=str(e))
    except search.ElasticSearchError as e:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='Elasticsearch could not process your query: %s' % str(e))


@router.post(
    '/query', tags=['entries/metadata'],
    summary='Search entries and retrieve their metadata',
    response_model=EntriesMetadataResponse,
    responses=create_responses(_bad_owner_response),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def post_entries_metadata_query(
        request: Request,
        data: EntriesMetadata,
        user: User = Depends(create_user_dependency())):

    '''
    Executes a *query* and returns a *page* of the results with *required* result data
    as well as *statistics* and *aggregated* data.

    This is the basic search operation to retrieve metadata for entries that match
    certain search criteria (`query` and `owner`). All parameters (including `query`, `owner`)
    are optional. Look at the body schema or parameter documentation for more details.

    By default the *empty* search (that returns everything) is performed. Only a small
    page of the search results are returned at a time; use `pagination` in subsequent
    requests to retrive more data. Each entry has a lot of different *metadata*, use
    `required` to limit the data that is returned.

    The `statistics` and `aggregations` keys will further allow to return statistics
    and aggregated data over all search results.
    '''

    return perform_search(
        owner=data.owner,
        query=data.query,
        pagination=data.pagination,
        required=data.required,
        statistics=data.statistics,
        aggregations=data.aggregations,
        user_id=user.user_id if user is not None else None)


@router.get(
    '', tags=[metadata_tag],
    summary='Search entries and retrieve their metadata',
    response_model=EntriesMetadataResponse,
    responses=create_responses(_bad_owner_response),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def get_entries_metadata(
        request: Request,
        with_query: WithQuery = Depends(query_parameters),
        pagination: EntryPagination = Depends(entry_pagination_parameters),
        required: MetadataRequired = Depends(metadata_required_parameters),
        user: User = Depends(create_user_dependency())):
    '''
    Executes a *query* and returns a *page* of the results with *required* result data.
    This is a version of `/entries/query`. Queries work a little different, because
    we cannot put complex queries into URL parameters.

    In addition to the `q` parameter (see parameter documentation for details), you can use all NOMAD
    search quantities as parameters, e.g. `?atoms=H&atoms=O`. Those quantities can be
    used with additional operators attached to their names, e.g. `?n_atoms__gte=3` for
    all entries with more than 3 atoms. Operators are `all`, `any`, `none`, `gte`,
    `gt`, `lt`, `lte`.
    '''

    res = perform_search(
        owner=with_query.owner, query=with_query.query,
        pagination=pagination, required=required,
        user_id=user.user_id if user is not None else None)
    res.pagination.populate_urls(request)
    return res


def _do_exaustive_search(owner: Owner, query: Query, include: List[str], user: User) -> Iterator[Dict[str, Any]]:
    page_after_value = None
    while True:
        response = perform_search(
            owner=owner, query=query,
            pagination=EntryPagination(size=100, page_after_value=page_after_value, order_by='upload_id'),
            required=MetadataRequired(include=include),
            user_id=user.user_id if user is not None else None)

        page_after_value = response.pagination.next_page_after_value

        for result in response.data:
            yield result

        if page_after_value is None or len(response.data) == 0:
            break


class _Uploads():
    '''
    A helper class that caches subsequent access to upload files the same upload.
    '''
    def __init__(self):
        self._upload_files = None

    def get_upload_files(self, upload_id: str) -> files.UploadFiles:
        if self._upload_files is not None and self._upload_files.upload_id != upload_id:
            self._upload_files.close()

        if self._upload_files is None or self._upload_files.upload_id != upload_id:
            self._upload_files = files.UploadFiles.get(
                upload_id, is_authorized=lambda *args, **kwargs: True)

        return self._upload_files

    def close(self):
        if self._upload_files is not None:
            self._upload_files.close()


def _create_entry_raw(entry_metadata: Dict[str, Any], uploads: _Uploads):
    calc_id = entry_metadata['calc_id']
    upload_id = entry_metadata['upload_id']
    mainfile = entry_metadata['mainfile']

    upload_files = uploads.get_upload_files(upload_id)
    mainfile_dir = os.path.dirname(mainfile)

    files = []
    for file_name, file_size in upload_files.raw_file_list(directory=mainfile_dir):
        path = os.path.join(mainfile_dir, file_name)
        files.append(EntryRawFile(path=path, size=file_size))

    return EntryRaw(calc_id=calc_id, upload_id=upload_id, mainfile=mainfile, files=files)


def _answer_entries_raw_request(
        owner: Owner, query: Query, pagination: EntryPagination, user: User):

    if owner == Owner.all_:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=strip('''
            The owner=all is not allowed for this operation as it will search for entries
            that you might now be allowed to access.
            '''))

    search_response = perform_search(
        owner=owner, query=query,
        pagination=pagination,
        required=MetadataRequired(include=['calc_id', 'upload_id', 'mainfile']),
        user_id=user.user_id if user is not None else None)

    uploads = _Uploads()
    try:
        response_data = [
            _create_entry_raw(entry_metadata, uploads)
            for entry_metadata in search_response.data]
    finally:
        uploads.close()

    return EntriesRawResponse(
        owner=search_response.owner,
        query=search_response.query,
        pagination=search_response.pagination,
        data=response_data)


def _answer_entries_raw_download_request(owner: Owner, query: Query, files: Files, user: User):
    if owner == Owner.all_:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=strip('''
            The owner=all is not allowed for this operation as it will search for entries
            that you might now be allowed to access.
            '''))

    response = perform_search(
        owner=owner, query=query,
        pagination=EntryPagination(page_size=0),
        required=MetadataRequired(include=[]),
        user_id=user.user_id if user is not None else None)

    if response.pagination.total > config.max_entry_download:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail='The limit of maximum number of entries in a single download (%d) has been exeeded (%d).' % (
                config.max_entry_download, response.pagination.total))

    uploads = _Uploads()
    files_params = Files() if files is None else files
    manifest = []
    search_includes = ['calc_id', 'upload_id', 'mainfile']
    streamed_paths: Set[str] = set()

    try:
        # a generator of File objects to create the streamed zip from
        def raw_file_generator():
            # go through all entries that match the query
            for entry_metadata in _do_exaustive_search(owner, query, include=search_includes, user=user):
                upload_id = entry_metadata['upload_id']
                mainfile = entry_metadata['mainfile']

                upload_files = uploads.get_upload_files(upload_id)
                mainfile_dir = os.path.dirname(mainfile)

                # go through all files that belong to this entry
                all_filtered = True
                files = upload_files.raw_file_list(directory=mainfile_dir)
                for file_name, file_size in files:
                    path = os.path.join(mainfile_dir, file_name)

                    # apply the filter
                    if files_params.re_pattern is not None and not files_params.re_pattern.search(path):
                        continue
                    all_filtered = False

                    # add upload_id to path used in streamed zip
                    streamed_path = os.path.join(upload_id, path)

                    # check if already streamed
                    if streamed_path in streamed_paths:
                        continue
                    streamed_paths.add(streamed_path)

                    # yield the file
                    with upload_files.raw_file(path, 'rb') as f:
                        yield File(path=streamed_path, f=f, size=file_size)

                if not all_filtered or len(files) == 0:
                    entry_metadata['mainfile'] = os.path.join(upload_id, mainfile)
                    manifest.append(entry_metadata)

            # add the manifest at the end
            manifest_content = json.dumps(manifest).encode()
            yield File(path='manifest.json', f=io.BytesIO(manifest_content), size=len(manifest_content))

        # create the streaming response with zip file contents
        content = create_streamed_zipfile(raw_file_generator(), compress=files_params.compress)
        return StreamingResponse(content, media_type='application/zip')
    except Exception as e:
        logger.error('exception while streaming download', exc_info=e)
    finally:
        uploads.close()


_entries_raw_query_docstring = strip('''
    Will perform a search and return a *page* of raw file metadata for entries fulfilling
    the query. This allows you to get a complete list of all rawfiles with their full
    path in their respective upload and their sizes. The first returned files for each
    entry, is their respective *mainfile*.

    Each entry on NOMAD represents a set of raw files. These are the input and output
    files (as well as additional auxiliary files) in their original form, i.e. as
    provided by the uploader. More specifically, an entry represents a code-run identified
    by a certain *mainfile*. This is usually the main output file of the code. All other
    files in the same directory are considered the entries *auxiliary* no matter their role
    or if they were actually parsed by NOMAD.

    This operation supports the usual `owner`, `query`, and `pagination` parameters.
    ''')


@router.post(
    '/raw/query',
    tags=[raw_tag],
    summary='Search entries and get their raw files metadata',
    description=_entries_raw_query_docstring,
    response_model=EntriesRawResponse,
    responses=create_responses(_bad_owner_response),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def post_entries_raw_query(
        request: Request, data: EntriesRaw, user: User = Depends(create_user_dependency())):

    return _answer_entries_raw_request(
        owner=data.owner, query=data.query, pagination=data.pagination, user=user)


@router.get(
    '/raw',
    tags=[raw_tag],
    summary='Search entries and get raw their raw files metadata',
    description=_entries_raw_query_docstring,
    response_model=EntriesRawResponse,
    response_model_exclude_unset=True,
    response_model_exclude_none=True,
    responses=create_responses(_bad_owner_response))
async def get_entries_raw(
        request: Request,
        with_query: WithQuery = Depends(query_parameters),
        pagination: EntryPagination = Depends(entry_pagination_parameters),
        user: User = Depends(create_user_dependency())):

    res = _answer_entries_raw_request(
        owner=with_query.owner, query=with_query.query, pagination=pagination, user=user)
    res.pagination.populate_urls(request)
    return res


_entries_raw_download_query_docstring = strip('''
    This operation will perform a search and stream a .zip file with raw input and output
    files of the found entries.

    Each entry on NOMAD represents a set of raw files. These are the input and output
    files (as well as additional auxiliary files) in their original form, i.e. as
    provided by the uploader. More specifically, an entry represents a code-run identified
    by a certain *mainfile*. This is usually the main output file of the code. All other
    files in the same directory are considered the entries *auxiliary* no matter their role
    or if they were actually parsed by NOMAD.

    After performing a search (that uses the same parameters as in all search operations),
    NOMAD will iterate through all results and create a .zip-file with all the entries'
    main and auxiliary files. The files will be organized in the same directory structure
    that they were uploaded in. The respective upload root directories are further prefixed
    with the `upload_id` of the respective uploads. The .zip-file will further contain
    a `manifest.json` with `upload_id`, `calc_id`, and `mainfile` of each entry.
    ''')


@router.post(
    '/raw/download/query',
    tags=[raw_tag],
    summary='Search entries and download their raw files',
    description=_entries_raw_download_query_docstring,
    response_class=StreamingResponse,
    responses=create_responses(_raw_download_response, _bad_owner_response))
async def post_entries_raw_download_query(
        data: EntriesRawDownload, user: User = Depends(create_user_dependency())):

    return _answer_entries_raw_download_request(
        owner=data.owner, query=data.query, files=data.files, user=user)


@router.get(
    '/raw/download',
    tags=[raw_tag],
    summary='Search entries and download their raw files',
    description=_entries_raw_download_query_docstring,
    response_class=StreamingResponse,
    responses=create_responses(_raw_download_response, _bad_owner_response))
async def get_entries_raw_download(
        with_query: WithQuery = Depends(query_parameters),
        files: Files = Depends(files_parameters),
        user: User = Depends(create_user_dependency())):

    return _answer_entries_raw_download_request(
        owner=with_query.owner, query=with_query.query, files=files, user=user)


def _read_archive(entry_metadata, uploads, required_reader: RequiredReader):
    calc_id = entry_metadata['calc_id']
    upload_id = entry_metadata['upload_id']
    upload_files = uploads.get_upload_files(upload_id)

    try:
        with upload_files.read_archive(calc_id) as archive:
            return {
                'calc_id': calc_id,
                'upload_id': upload_id,
                'parser_name': entry_metadata['parser_name'],
                'archive': required_reader.read(archive, calc_id)
            }
    except ArchiveQueryError as e:
        raise HTTPException(status_code=status.HTTP_400_BAD_REQUEST, detail=str(e))


def _validate_required(required: ArchiveRequired) -> RequiredReader:
    try:
        return RequiredReader(required)
    except RequiredValidationError as e:
        raise HTTPException(status_code=status.HTTP_422_UNPROCESSABLE_ENTITY, detail=[dict(
            msg=e.msg, loc=['required'] + e.loc)])


def _answer_entries_archive_request(
        owner: Owner, query: Query, pagination: EntryPagination, required: ArchiveRequired,
        user: User):

    if owner == Owner.all_:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=strip('''
            The owner=all is not allowed for this operation as it will search for entries
            that you might now be allowed to access.
            '''))

    if required is None:
        required = '*'

    required_reader = _validate_required(required)

    search_response = perform_search(
        owner=owner, query=query,
        pagination=pagination,
        required=MetadataRequired(include=['calc_id', 'upload_id', 'parser_name']),
        user_id=user.user_id if user is not None else None)

    uploads = _Uploads()
    response_data = {}
    for entry_metadata in search_response.data:
        calc_id, upload_id = entry_metadata['calc_id'], entry_metadata['upload_id']

        archive_data = None
        try:
            archive_data = _read_archive(entry_metadata, uploads, required_reader)['archive']
        except KeyError as e:
            logger.error('missing archive', exc_info=e, calc_id=calc_id)
            continue

        response_data[calc_id] = {
            'calc_id': calc_id,
            'upload_id': upload_id,
            'parser_name': entry_metadata['parser_name'],
            'archive': archive_data}

    uploads.close()

    return EntriesArchiveResponse(
        owner=search_response.owner,
        query=search_response.query,
        pagination=search_response.pagination,
        required=required,
        data=list(response_data.values()))


_entries_archive_docstring = strip('''
    This operation will perform a search with the given `query` and `owner` and return
    the a *page* of `required` archive data. Look at the body schema or parameter documentation
    for more details. The **GET** version of this operation will only allow to provide
    the full archives.
    ''')


@router.post(
    '/archive/query',
    tags=[archive_tag],
    summary='Search entries and access their archives',
    description=_entries_archive_docstring,
    response_model=EntriesArchiveResponse,
    response_model_exclude_unset=True,
    response_model_exclude_none=True,
    responses=create_responses(_bad_owner_response, _bad_archive_required_response))
async def post_entries_archive_query(
        request: Request, data: EntriesArchive, user: User = Depends(create_user_dependency())):

    return _answer_entries_archive_request(
        owner=data.owner, query=data.query, pagination=data.pagination,
        required=data.required, user=user)


@router.get(
    '/archive',
    tags=[archive_tag],
    summary='Search entries and access their archives',
    description=_entries_archive_docstring,
    response_model=EntriesArchiveResponse,
    response_model_exclude_unset=True,
    response_model_exclude_none=True,
    responses=create_responses(_bad_owner_response, _bad_archive_required_response))
async def get_entries_archive_query(
        request: Request,
        with_query: WithQuery = Depends(query_parameters),
        pagination: EntryPagination = Depends(entry_pagination_parameters),
        user: User = Depends(create_user_dependency())):

    res = _answer_entries_archive_request(
        owner=with_query.owner, query=with_query.query, pagination=pagination,
        required=None, user=user)
    res.pagination.populate_urls(request)
    return res


def _answer_entries_archive_download_request(
        owner: Owner, query: Query, files: Files, user: User):

    if owner == Owner.all_:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail=strip('''
            The owner=all is not allowed for this operation as it will search for entries
            that you might now be allowed to access.
            '''))

    files_params = Files() if files is None else files

    response = perform_search(
        owner=owner, query=query,
        pagination=EntryPagination(page_size=0),
        required=MetadataRequired(include=[]),
        user_id=user.user_id if user is not None else None)

    if response.pagination.total > config.max_entry_download:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=(
                'The limit of maximum number of entries in a single download (%d) has been '
                'exeeded (%d).' % (config.max_entry_download, response.pagination.total)))

    uploads = _Uploads()
    manifest = []
    search_includes = ['calc_id', 'upload_id', 'parser_name']

    required_reader = RequiredReader('*')

    # a generator of File objects to create the streamed zip from
    def file_generator():
        # go through all entries that match the query
        for entry_metadata in _do_exaustive_search(owner, query, include=search_includes, user=user):
            path = os.path.join(entry_metadata['upload_id'], '%s.json' % entry_metadata['calc_id'])
            try:
                archive_data = _read_archive(entry_metadata, uploads, required_reader)

                f = io.BytesIO(orjson.dumps(
                    archive_data, option=orjson.OPT_INDENT_2 | orjson.OPT_NON_STR_KEYS))

                yield File(path=path, f=f, size=f.getbuffer().nbytes)
            except KeyError as e:
                logger.error('missing archive', calc_id=entry_metadata['calc_id'], exc_info=e)

            entry_metadata['path'] = path
            manifest.append(entry_metadata)

        # add the manifest at the end
        manifest_content = json.dumps(manifest).encode()
        yield File(path='manifest.json', f=io.BytesIO(manifest_content), size=len(manifest_content))

    try:
        # create the streaming response with zip file contents
        content = create_streamed_zipfile(file_generator(), compress=files_params.compress)
        return StreamingResponse(content, media_type='application/zip')
    finally:
        uploads.close()


_entries_archive_download_docstring = strip('''
    This operation will perform a search with the given `query` and `owner` and stream
    a .zip-file with the full archive contents for all matching entries. This is not
    paginated. Look at the body schema or parameter documentation for more details.
    ''')


@router.post(
    '/archive/download/query',
    tags=[archive_tag],
    summary='Search entries and download their archives',
    description=_entries_archive_download_docstring,
    response_class=StreamingResponse,
    responses=create_responses(
        _archive_download_response, _bad_owner_response, _bad_archive_required_response))
async def post_entries_archive_download_query(
        data: EntriesArchiveDownload, user: User = Depends(create_user_dependency())):

    return _answer_entries_archive_download_request(
        owner=data.owner, query=data.query, files=data.files, user=user)


@router.get(
    '/archive/download',
    tags=[archive_tag],
    summary='Search entries and download their archives',
    description=_entries_archive_download_docstring,
    response_class=StreamingResponse,
    responses=create_responses(
        _archive_download_response, _bad_owner_response, _bad_archive_required_response))
async def get_entries_archive_download(
        with_query: WithQuery = Depends(query_parameters),
        files: Files = Depends(files_parameters),
        user: User = Depends(create_user_dependency())):

    return _answer_entries_archive_download_request(
        owner=with_query.owner, query=with_query.query, files=files, user=user)


@router.get(
    '/{entry_id}', tags=[metadata_tag],
    summary='Get the metadata of an entry by its id',
    response_model=EntryMetadataResponse,
    responses=create_responses(_bad_id_response),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def get_entry_metadata(
        entry_id: str = Path(..., description='The unique entry id of the entry to retrieve metadata from.'),
        required: MetadataRequired = Depends(metadata_required_parameters),
        user: User = Depends(create_user_dependency())):
    '''
    Retrives the entry metadata for the given id.
    '''

    query = {'calc_id': entry_id}
    response = perform_search(owner=Owner.all_, query=query, required=required, user_id=user.user_id if user is not None else None)

    if response.pagination.total == 0:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail='The entry with the given id does not exist or is not visible to you.')

    return {
        'entry_id': entry_id,
        'required': required,
        'data': response.data[0]
    }


@router.get(
    '/{entry_id}/raw',
    tags=[raw_tag],
    summary='Get the raw files metadata for an entry by its id',
    response_model=EntryRawResponse,
    responses=create_responses(_bad_id_response),
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def get_entry_raw(
        entry_id: str = Path(..., description='The unique entry id of the entry to retrieve raw data from.'),
        files: Files = Depends(files_parameters),
        user: User = Depends(create_user_dependency())):
    '''
    Returns the file metadata for all input and output files (including auxiliary files)
    of the given `entry_id`. The first file will be the *mainfile*.
    '''
    query = dict(calc_id=entry_id)
    response = perform_search(
        owner=Owner.visible, query=query,
        required=MetadataRequired(include=['calc_id', 'upload_id', 'mainfile']),
        user_id=user.user_id if user is not None else None)

    if response.pagination.total == 0:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail='The entry with the given id does not exist or is not visible to you.')

    uploads = _Uploads()
    try:
        return EntryRawResponse(entry_id=entry_id, data=_create_entry_raw(response.data[0], uploads))
    finally:
        uploads.close()


@router.get(
    '/{entry_id}/raw/download',
    tags=[raw_tag],
    summary='Get the raw data of an entry by its id',
    response_class=StreamingResponse,
    responses=create_responses(_bad_id_response, _raw_download_response))
async def get_entry_raw_download(
        entry_id: str = Path(..., description='The unique entry id of the entry to retrieve raw data from.'),
        files: Files = Depends(files_parameters),
        user: User = Depends(create_user_dependency())):
    '''
    Streams a .zip file with the raw files from the requested entry.
    '''
    query = dict(calc_id=entry_id)
    response = perform_search(
        owner=Owner.visible, query=query,
        required=MetadataRequired(include=['calc_id']),
        user_id=user.user_id if user is not None else None)

    if response.pagination.total == 0:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail='The entry with the given id does not exist or is not visible to you.')

    return _answer_entries_raw_download_request(owner=Owner.public, query=query, files=files, user=user)


class FileContentIterator:
    '''
    An iterator implementation that provides the contents of an underlying file, based on
    offset and length.

    Arguments:
        f: the file-like
        offset: the offset
        length: the amount of bytes
    '''
    def __init__(self, f, offset, length):
        self.f = f
        self.offset = offset
        self.read_bytes = 0
        self.f.seek(self.offset)
        self.length = length

    def __next__(self):
        remaining = self.length - self.read_bytes
        if remaining > 0:
            content = self.f.read(remaining)
            content_length = len(content)
            self.read_bytes += content_length
            if content_length == 0:
                self.length = self.read_bytes
            return content
        else:
            raise StopIteration


@router.get(
    '/{entry_id}/raw/download/{path}',
    tags=[raw_tag],
    summary='Get the raw data of an entry by its id',
    response_class=StreamingResponse,
    responses=create_responses(_bad_id_response, _bad_path_response, _raw_download_file_response))
async def get_entry_raw_download_file(
        entry_id: str = Path(..., description='The unique entry id of the entry to retrieve raw data from.'),
        path: str = Path(..., description='A relative path to a file based on the directory of the entry\'s mainfile.'),
        offset: Optional[int] = QueryParameter(
            0, ge=0, description=strip('''
                Integer offset that marks the start of the contents to retrieve. Default
                is the start of the file.''')),
        length: Optional[int] = QueryParameter(
            -1, ge=0, description=strip('''
                The amounts of contents in bytes to stream. By default, the remainder of
                the file is streamed.''')),
        decompress: Optional[bool] = QueryParameter(
            False, description=strip('''
                Attempt to decompress the contents, if the file is .gz or .xz.''')),
        user: User = Depends(create_user_dependency())):
    '''
    Streams the contents of an individual file from the requested entry.
    '''
    query = dict(calc_id=entry_id)
    response = perform_search(
        owner=Owner.visible, query=query,
        required=MetadataRequired(include=['calc_id', 'upload_id', 'mainfile']),
        user_id=user.user_id if user is not None else None)

    if response.pagination.total == 0:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail='The entry with the given id does not exist or is not visible to you.')

    entry_metadata = response.data[0]
    upload_id, mainfile = entry_metadata['upload_id'], entry_metadata['mainfile']
    # The user is allowed to access all files, because the entry is in the "visible" scope
    upload_files = files.UploadFiles.get(upload_id, is_authorized=lambda *args, **kwargs: True)

    entry_path = os.path.dirname(mainfile)
    path = os.path.join(entry_path, path)

    raw_file: Any = None
    try:
        raw_file = upload_files.raw_file(path, 'br')

        if decompress:
            if path.endswith('.gz'):
                raw_file = gzip.GzipFile(filename=path[:3], mode='rb', fileobj=raw_file)

            if path.endswith('.xz'):
                raw_file = lzma.open(filename=raw_file, mode='rb')

        # We only provide a specific mime-type, if the whole file is requested. Otherwise,
        # it is unlikely that the provided contents will match the overall file mime-type.
        mime_type = 'application/octet-stream'
        if offset == 0 and length < 0:
            buffer = raw_file.read(2048)
            raw_file.seek(0)
            mime_type = magic.from_buffer(buffer, mime=True)

        raw_file_content: Union[FileContentIterator, IO] = None
        if length > 0:
            raw_file_content = FileContentIterator(raw_file, offset, length)
        else:
            raw_file.seek(offset)
            raw_file_content = raw_file

        return StreamingResponse(raw_file_content, media_type=mime_type)

    except KeyError:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail='The requested file does not exist.')


def _answer_entry_archive_request(entry_id: str, required: ArchiveRequired, user: User):
    required_reader = _validate_required(required)

    query = dict(calc_id=entry_id)
    response = perform_search(
        owner=Owner.visible, query=query,
        required=MetadataRequired(include=['calc_id', 'upload_id', 'parser_name']),
        user_id=user.user_id if user is not None else None)

    if response.pagination.total == 0:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail='The entry with the given id does not exist or is not visible to you.')

    entry_metadata = response.data[0]

    uploads = _Uploads()
    try:
        try:
            archive_data = _read_archive(entry_metadata, uploads, required_reader)['archive']
        except KeyError:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND,
                detail='The entry with the given id does exist, but it has no archive.')

        return {
            'entry_id': entry_id,
            'required': required,
            'data': {
                'calc_id': entry_id,
                'upload_id': entry_metadata['upload_id'],
                'parser_name': entry_metadata['parser_name'],
                'archive': archive_data
            }}
    finally:
        uploads.close()


@router.get(
    '/{entry_id}/archive',
    tags=[archive_tag],
    summary='Get the archive for an entry by its id',
    response_model=EntryArchiveResponse,
    response_model_exclude_unset=True,
    response_model_exclude_none=True,
    responses=create_responses(_bad_id_response))
async def get_entry_archive(
        entry_id: str = Path(..., description='The unique entry id of the entry to retrieve raw data from.'),
        user: User = Depends(create_user_dependency())):
    '''
    Returns the full archive for the given `entry_id`.
    '''
    return _answer_entry_archive_request(entry_id=entry_id, required='*', user=user)


@router.post(
    '/{entry_id}/archive/query',
    tags=[archive_tag],
    summary='Get the archive for an entry by its id',
    response_model=EntryArchiveResponse,
    response_model_exclude_unset=True,
    response_model_exclude_none=True,
    responses=create_responses(_bad_id_response, _bad_archive_required_response))
async def post_entry_archive_query(
        data: EntryArchiveRequest, user: User = Depends(create_user_dependency()),
        entry_id: str = Path(..., description='The unique entry id of the entry to retrieve raw data from.')):

    '''
    Returns a partial archive for the given `entry_id` based on the `required` specified
    in the body.
    '''
    return _answer_entry_archive_request(entry_id=entry_id, required=data.required, user=user)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Any, Tuple, Dict, Union, List

from nomad import infrastructure, config
from nomad.metainfo import MSection, Definition, Quantity, Reference, SubSection, Section
from nomad.datamodel import EntryArchive
from nomad.datamodel.metainfo.common import FastAccess


def create_partial_archive(archive: EntryArchive) -> Dict:
    '''
    Creates a partial archive JSON serializable dict that can be stored directly.
    The given archive is filtered based on the metainfo category ``FastAccess``.
    Selected sections and other data that they reference (recursively) comprise the
    resulting partial archive.

    TODO at the moment is hard coded and NOT informed by the metainfo. We simply
    add sections EntryMetadata and Workflow.

    Arguments:
        archive: The archive as an :class:`EntryArchive` instance.

    Returns: the partial archive in JSON serializable dictionary form.
    '''
    # A list with all referenced sections that might not yet been ensured to be in the
    # resulting partial archive
    referenceds: List[MSection] = []
    # contents keeps track of all sections in the partial archive by keeping their
    # JSON serializable form and placeholder status in a dict
    contents: Dict[MSection, Tuple[dict, bool]] = dict()

    def partial(definition: Definition, section: MSection) -> bool:
        '''
        ``m_to_dict`` partial function that selects what goes into the partial archive
        and what not. It also collects references as a side-effect.
        '''
        if section.m_def == EntryArchive.m_def:
            if definition.m_def == Quantity:
                return True
            return FastAccess.m_def in definition.categories

        if isinstance(definition, Quantity) and isinstance(definition.type, Reference) \
                and FastAccess.m_def in definition.categories:
            # Reference list in partial archives are not supported
            if definition.is_scalar:
                referenced = getattr(section, definition.name)
                if referenced is not None and referenced not in contents:
                    referenceds.append(referenced)

        if isinstance(definition, SubSection):
            return FastAccess.m_def in definition.categories

        return True

    # add the main content
    partial_contents = archive.m_to_dict(partial=partial)

    # add the referenced data
    def add(section, placeholder=False) -> dict:
        '''
        Adds the given section to partial_contents at the right place. If not a placeholder,
        the section's serialization is added (or replacing an existing placeholder).
        Otherwise, an empty dict is added as a placeholder for referenced children.
        '''
        result: Dict[str, Any] = None
        content, content_is_placeholder = contents.get(section, (None, True))
        if content is not None:
            if content_is_placeholder and not placeholder:
                # the placeholder gets replaced later
                pass
            else:
                return content

        if section.m_parent is None:
            contents[section] = partial_contents, False
            return partial_contents

        parent_dict = add(section.m_parent, placeholder=True)
        if placeholder:
            result = {}
        else:
            result = section.m_to_dict(partial=partial)

        sub_section = section.m_parent_sub_section
        if sub_section.repeats:
            sections = parent_dict.setdefault(sub_section.name, [])
            while len(sections) < section.m_parent_index + 1:
                sections.append(None)
            sections[section.m_parent_index] = result
        else:
            parent_dict[sub_section.name] = result

        contents[section] = result, placeholder
        return result

    # we add referenced objects as long as they are added by subsequent serialization
    # of referenced sections to implement the recursive nature of further references in
    # already referenced sections.
    while len(referenceds) > 0:
        referenced = referenceds.pop()
        add(referenced)

    return partial_contents


def write_partial_archive_to_mongo(archive: EntryArchive):
    ''' Partially writes the given archive to mongodb. '''
    mongo_db = infrastructure.mongo_client[config.mongo.db_name]
    mongo_collection = mongo_db['archive']
    mongo_id = archive.section_metadata.calc_id

    partial_archive_dict = create_partial_archive(archive)
    partial_archive_dict['_id'] = mongo_id
    mongo_collection.replace_one(dict(_id=mongo_id), partial_archive_dict, upsert=True)


def read_partial_archive_from_mongo(entry_id: str, as_dict=False) -> Union[EntryArchive, Dict]:
    '''
    Reads the partial archive for the given id from mongodb.

    Arguments:
        entry_id: The entry id for the entry.
        as_dict: Return the JSON serializable dictionary form of the archive not the
            :class:`EntryArchive` form.
    '''
    mongo_db = infrastructure.mongo_client[config.mongo.db_name]
    mongo_collection = mongo_db['archive']
    archive_dict = mongo_collection.find_one(dict(_id=entry_id))

    if as_dict:
        return archive_dict

    return EntryArchive.m_from_dict(archive_dict)


def delete_partial_archives_from_mongo(entry_ids: List[str]):
    mongo_db = infrastructure.mongo_client[config.mongo.db_name]
    mongo_collection = mongo_db['archive']
    mongo_collection.delete_many(dict(_id={'$in': entry_ids}))


def read_partial_archives_from_mongo(entry_ids: List[str], as_dict=False) -> Dict[str, Union[EntryArchive, Dict]]:
    '''
    Reads the partial archives for a set of entries.

    Arguments:
        entry_ids: A list of entry ids.
        as_dict: Return the JSON serializable dictionary form of the archive not the
            :class:`EntryArchive` form.

    Returns:
        A dictionary with entry_ids as keys.
    '''
    mongo_db = infrastructure.mongo_client[config.mongo.db_name]
    mongo_collection = mongo_db['archive']
    archive_dicts = mongo_collection.find(dict(_id={'$in': entry_ids}))

    if as_dict:
        return {archive_dict.pop('_id'): archive_dict for archive_dict in archive_dicts}

    return {
        archive_dict.pop('_id'): EntryArchive.m_from_dict(archive_dict)
        for archive_dict in archive_dicts}


__all_parent_sections: Dict[Section, Tuple[str, Section]] = {}


def _all_parent_sections():
    if len(__all_parent_sections) == 0:
        def add(section):
            for sub_section in section.all_sub_sections.values():
                sub_section_section = sub_section.sub_section.m_resolved()
                __all_parent_sections.setdefault(sub_section_section, []).append((sub_section.qualified_name(), section, ))
                add(sub_section_section)

        add(EntryArchive.m_def)

    return __all_parent_sections


class _Incomplete(Exception): pass


def compute_required_with_referenced(required):
    '''
    Updates the given required dictionary to ensure that references to non required
    sections within a partial fast access archive are included. Only references that
    are directly contained in required are added. References from wildcarded sub sections
    are ignored.

    Returns: A new required dict or None. None is returned if it is unclear if the required
    is only accessing information of fast access partial archives.
    '''
    # TODO this function should be based on the metainfo

    if not isinstance(required, dict):
        return None

    if any(key.startswith('section_run') for key in required):
        return None

    required = dict(**required)

    def add_parent_section(section, child_required):
        parent_sections = _all_parent_sections().get(section, [])
        if len(parent_sections) == 0:
            return [required]

        result = []
        for name, parent_section in parent_sections:
            child_key = name.split('.')[-1]
            for parent_required in add_parent_section(parent_section, None):
                result.append(parent_required.setdefault(child_key, child_required if child_required else {}))

        return result

    def traverse(
            current: Union[dict, str],
            parent: Section = EntryArchive.m_def):

        if isinstance(current, str):
            return

        current_updates = {}
        for key, value in current.items():
            prop = key.split('[')[0]
            prop_definition = parent.all_properties[prop]
            if isinstance(prop_definition, SubSection):
                if FastAccess.m_def not in prop_definition.categories:
                    raise _Incomplete()

                traverse(value, prop_definition.sub_section)
            if isinstance(prop_definition, Quantity) and isinstance(prop_definition.type, Reference):
                current_updates[prop] = '*'
                if FastAccess.m_def not in prop_definition.categories:
                    continue

                target_section_def = prop_definition.type.target_section_def.m_resolved()
                # TODO is this a bug, should the result of the call not be used to traverse?
                add_parent_section(target_section_def, value)
                traverse(value, target_section_def)
        current.update(**current_updates)

    try:
        traverse(dict(**required))
    except _Incomplete:
        # We realized that something is required that is not in the partial archive
        return None

    return required



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
Automatically synchronizes nomad it with a given database. It creates a list of paths
to mainfiles in nomad and compares it with paths in the external database. The missing
paths in nomad will then be downloaded from the external database and subsequently
uploaded to nomad. The downloaded files are by default saved in '/nomad/fairdi/external'.
'''

import requests
import re
import subprocess
from urllib import parse as urllib_parse
import os
import datetime
import click
import tarfile
import threading
import time
import typing

from .client import client
from nomad import config
from nomad.cli.client import upload as nomad_upload


class DbUpdater:
    def __init__(self, *args, **kwargs):
        self.db_name = 'aflowlib'
        self.root_url = 'http://aflowlib.duke.edu/AFLOWDATA/LIB1_LIB'
        self.local_path = '/nomad/fairdi/external'
        self.dbfile = None
        self.nomadfile = None
        self.outfile = None
        self.do_download = False
        self.do_upload = False
        self.do_publish = False
        self.cleanup = False
        self.parallel = 2
        self.max_depth = 10
        self.uids = []
        self._set(**kwargs)
        self._set_db()
        self._set_local_path()
        self.configure_client()

    def _set(self, **kwargs):
        for key, val in kwargs.items():
            key = key.lower()
            if hasattr(self, key):
                setattr(self, key, val)
            else:
                raise KeyError('Invalid key %s' % key)

    def configure_client(self, **kwargs):
        from nomad.cli.client import create_client
        self.client = create_client()

    def _set_local_path(self):
        subdir = ''
        if self.db_name.lower() == 'aflowlib':
            subdir = self.root_url.strip('/').split('/')[-1]
        dbpath = os.path.join(self.local_path, self.db_name)
        self._local_path = os.path.join(dbpath, subdir)
        if not os.path.isdir(dbpath):
            os.mkdir(dbpath)
        if not os.path.isdir(self._local_path):
            os.mkdir(self._local_path)

    def _set_db(self):
        if self.db_name.lower() == 'aflowlib':
            self.max_depth = 4
        else:
            raise NotImplementedError('%s not yet supported.' % self.db_name)

    def _open_page(self, path: str):
        with requests.Session() as session:
            response = session.get(path, verify=False)
        if not response.ok:
            return response.raise_for_status()
        return response

    def get_paths(self, root: str) -> typing.List[str]:
        response = self._open_page(root)
        paths = []
        for url in re.findall('<a href="([^"]+)">', str(response.content)):
            if url in response.url:
                continue
            paths.append(os.path.join(response.url, url.lstrip('./')))
        return paths

    def _is_mainfile(self, path: str) -> bool:
        if 'vasprun.xml' in path:
            return True
        return False

    def _is_dir(self, path: str) -> bool:
        path = path.strip()
        if path[-1] == '/' and self.root_url in path:
            return True
        return False

    def _depth(self, path: str) -> int:
        return urllib_parse.urlparse(path).path.rstrip('/').count('/')

    def _rules_ok(self, path: str) -> bool:
        ok = True
        if '?' in path:
            ok = False
        return ok

    def _to_namesafe(self, path: str) -> str:
        path = path.strip('/')
        return re.sub(r'[^\w\d-]', '_', path)

    def _read_from_file(self, filename):
        print('Reading from file %s' % filename)
        data = []
        with open(filename) as f:
            line = f.readline()
            while line:
                data_i = line.strip().split()

                if len(data_i) == 1:
                    data.append(data_i[0])

                else:
                    info = data_i[1] in ['T', 'True', '1', 'Y']
                    data.append((data_i[0], info))

                line = f.readline()
        return data

    def _write_to_file(self, data: typing.List, filename: str):
        with open(filename, 'w') as f:
            for i in range(len(data)):
                if isinstance(data[i], str):
                    f.write('%s\n' % data[i])
                else:
                    f.write('%s %s \n' % (data[i][0], data[i][1]))

    def _gen_db_list(self):
        print('Generating list from %s' % self.root_url)
        self.db_files = []
        todo = self.get_paths(self.root_url)
        while len(todo) > 0:
            cur = todo.pop(0)
            if not self._is_dir(cur):
                continue

            if self._depth(cur) == self.max_depth:
                self.db_files.append(cur)

            else:
                add = self.get_paths(cur)
                add = [path for path in add if self._rules_ok(path)]
                todo = add + todo

        if self.dbfile is not None:
            self._write_to_file(self.db_files, self.dbfile)

    def get_db_list(self):
        if self.dbfile is not None and os.path.isfile(self.dbfile):
            self.db_files = self._read_from_file(self.dbfile)
        else:
            self._gen_db_list()

    def _gen_nomad_list(self):
        print('Generating NOMAD list')
        if self.db_name.lower() == 'aflowlib':
            servers = ['LIB%d_LIB' % n for n in range(1, 10)]
            paths = []
            for s in servers:
                if s in self.root_url:
                    paths.append(s)
            if len(paths) == 0:
                paths = servers
            # uploader: Stefano Curtarolo
            kwargs = dict(
                uploader_id='81b96683-7170-49d7-8c4e-e9f34906b3ea',
                paths=paths,
                scroll=True)

        self.nomad_files = []
        while True:
            res = self.client.repo.search(**kwargs).response()
            results = res.result.results
            if len(results) == 0:
                break
            for i in range(len(results)):
                self.nomad_files.append(results[i]['mainfile'])
            scroll_id = res.result.scroll.scroll_id
            kwargs.update({'scroll_id': scroll_id})

        if self.nomadfile is not None:
            self._write_to_file(self.nomad_files, self.nomadfile)

    def get_nomad_list(self):
        if self.nomadfile is not None and os.path.isfile(self.nomadfile):
            self.nomad_files = self._read_from_file(self.nomadfile)
        else:
            self._gen_nomad_list()

    def _reduce_list(self, ilist: typing.List[str]):
        olist = []
        for e in ilist:
            p = urllib_parse.urlparse(e).path.strip('/')
            olist.append(os.path.join(*p.split('/')[1:self.max_depth]))
        olist = list(set(olist))
        olist.sort()
        return olist

    def compare_lists(self):
        print('Identifying differences')
        db = self._reduce_list(self.db_files)
        nomad = self._reduce_list(self.nomad_files)
        ns = set(nomad)
        self.update_list = [i for i in db if i not in ns]

        ds = set(db)
        self.in_nomad = [i for i in nomad if i not in ds]
        if len(self.in_nomad) > 0:
            fn = 'in_nomad.txt'
            print('Warning: Some NOMAD entries not found in db.')
            print('See %s for list.' % fn)
            self._write_to_file(self.in_nomad, fn)

        # add the root back
        u = urllib_parse.urlparse(self.root_url)
        up = u.path.strip('/').split('/')[0]
        root = '%s://%s/%s' % (u.scheme, u.netloc, up)
        self.update_list = [os.path.join(root, e) for e in self.update_list]
        self.is_updated_list = [False] * len(self.update_list)
        print('Found %d entries to be added in NOMAD' % len(self.update_list))

        if self.outfile is not None:
            data = [self.update_list[i] for i in range(len(self.update_list))]
            self._write_to_file(data, self.outfile)

    def _download(self, path: str, iodir: str) -> float:
        files = self.get_paths(path)
        files = [f for f in files if self._rules_ok(f)]
        size = 0.0
        with requests.Session() as session:
            for f in files:
                if self._is_dir(f):
                    _, s = self.get_files(f)
                    size += s
                    continue

                res = session.get(f, stream=True)
                fn = res.url.split('/')[-1]
                fn = os.path.join(iodir, fn)
                with open(fn, 'wb') as fb:
                    fb.write(res.content)

                size += os.path.getsize(fn)
        return size

    def _download_size(self, dirname: str) -> float:
        complete = False
        files = os.listdir(dirname)

        if self.db_name.lower() == 'aflowlib':
            if 'vasprun.xml.relax2.xz' in files:
                complete = True
        else:
            complete = True

        size = 0.0
        if not complete:
            self._cleanup([os.path.join(dirname, f) for f in files])
        else:
            for f in files:
                size += os.path.getsize(os.path.join(dirname, f))

        return size

    def get_files(self, path: str) -> typing.Tuple[str, float]:
        dirname = urllib_parse.urlparse(path).path
        dirname = self._to_namesafe(dirname)
        dirname = os.path.join(self._local_path, dirname)

        if os.path.isdir(dirname):
            size = self._download_size(dirname)
            if size == 0.0:
                size = self._download(path, dirname)

        else:
            os.mkdir(dirname)
            size = self._download(path, dirname)

        return dirname, size

    def _tar_files(self, dirs: typing.List[str], tarname: str):
        if os.path.isfile(tarname):
            return

        try:
            f = tarfile.open(tarname, 'w')
            for d in dirs:
                files = os.listdir(d)
                for fn in files:
                    f.add(os.path.join(d, fn))
            f.close()

        except Exception as e:
            os.remove(tarname)
            print('Error writing tar file %s. %s' % (tarname, e))

    def _make_name(self, dirs: typing.List[str]) -> typing.Tuple[str, str]:
        # name will be first and last entries
        d1 = self._to_namesafe(dirs[0].lstrip(self._local_path))
        d2 = self._to_namesafe(dirs[-1].lstrip(self._local_path))

        tarname = '%s-%s' % (d1, d2)
        uploadname = 'AFLOWLIB_%s' % tarname
        tarname = os.path.join(self._local_path, tarname + '.tar')

        return tarname, uploadname

    def _cleanup(self, ilist: typing.Union[str, typing.List[str]]):
        if isinstance(ilist, str):
            ilist = [ilist]
        for name in ilist:
            subprocess.Popen(['rm', '-rf', name])

    def _is_done_upload(self, uid: int) -> bool:
        res = self.client.uploads.get_upload(upload_id=uid).response().result
        Nproc = res.processed_calcs
        Ncalc = res.total_calcs

        if Nproc != Ncalc:
            return False
        return True

    def _get_status_upload(self, uploadname: str) -> typing.Tuple[str, int]:
        res = self.client.uploads.get_uploads(name=uploadname, state='all').response().result
        entries = res.results
        status = None
        upload_id = None

        for entry in entries:
            if entry['name'] == uploadname:
                status = 'uploaded'
                if entry['published']:
                    status = 'published'
                upload_id = entry['upload_id']
                break

        return status, upload_id

    def get_payload(self, uid: int) -> typing.Dict[str, typing.Any]:
        timenow = datetime.datetime.utcnow()
        if self.db_name == 'aflowlib':
            return dict(
                operation='publish',
                metadata=dict(
                    with_embargo=False,
                    comment='',
                    references=[
                        'http://www.sciencedirect.com/science/article/pii/S0927025614003322',
                        'http://aflowlib.org',
                        'http://www.sciencedirect.com/science/article/pii/S0927025612000687'],
                    coauthors=[
                        'f409d859-2639-4f82-b198-85e1c7c62f8b',
                        '580f7036-97b8-42b1-a9e6-815058bcac72',
                        '4c8d767d-335f-4ccd-9459-d0152b2026bc',
                        '7effd16a-a65c-4d95-b692-652488d94146',
                        'd59b3610-6335-4ad8-aca0-24a905de3a25',
                        '3df68fed-6ca0-4bf9-b2b1-d6be71e18f72',
                        'f7b540eb-a266-4379-9611-911ed8e3630e',
                        '7b3fe468-0011-4ba8-bd53-1ee65acda114',
                        '48e7a028-1a41-440d-9986-38540f5079c9',
                        'd63d07e6-ccc8-4eac-82af-d841627b6c53',
                        '9c308f66-eed1-4996-b33e-af78fb4944c7',
                        'd2621bc7-c45a-4d35-9dc1-5c05fa8326cb',
                        'ecba0e68-65ee-4b40-8fbf-a42714b1072b',
                        '81b96683-7170-49d7-8c4e-e9f34906b3ea'],
                    shared_with=[],
                    _upload_time=timenow,
                    _uploader='81b96683-7170-49d7-8c4e-e9f34906b3ea'))
        else:
            raise NotImplementedError('%s not yet supported.' % self.db_name)

    def publish(self, uids=None):
        print('Publishing')
        if uids is None:
            uids = self.uids
        for uid in uids:
            if self._is_done_upload(uid):
                payload = self.get_payload(uid)
                self.client.uploads.exec_upload_operation(upload_id=uid, payload=payload).response()

    def upload(self, file_path: str, name: str) -> int:
        res = self.client.uploads.upload(
            local_path=os.path.abspath(file_path), name=name).response().result
        return res.upload_id

    def download_proc(self, plist: typing.List[str]):
        size = 0.0
        max_zip_size = config.max_upload_size
        dirs = []
        done = []
        for i in range(len(plist)):
            d, s = self.get_files(self.update_list[plist[i]])
            if not self.do_upload:
                continue

            size += s
            dirs.append(d)
            done.append(plist[i])
            if size > max_zip_size or i == (len(plist) - 1):
                tarname, uploadname = self._make_name(dirs)
                status, uid = self._get_status_upload(uploadname)
                if status == 'published':
                    continue
                if status != 'uploaded':
                    self._tar_files(dirs, tarname)
                    uid = nomad_upload.upload_file(tarname, name=uploadname, offline=True)
                if self.do_publish:
                    self.publish([uid])
                self.uids.append(uid)
                if self.cleanup:
                    self._cleanup(dirs)
                    self._cleanup(tarname)
                size = 0.0
                dirs = []

    def download(self):
        print('Downloading from %s' % self.root_url)
        s = time.time()
        plist = [[] for i in range(self.parallel)]
        cur = 0
        for i in range(len(self.update_list)):
            if self.is_updated_list[i]:
                continue
            plist[cur % self.parallel].append(i)
            cur += 1

        if self.parallel > 1:
            procs = []
            for i in range(self.parallel):
                p = threading.Thread(target=self.download_proc, args=(plist[i],))
                procs.append(p)
            [p.start() for p in procs]
            [p.join() for p in procs]

        else:
            self.download_proc(plist[0])

        print('Time for download and upload (s)', time.time() - s)

    def prep_list(self):
        if self.outfile is not None and os.path.isfile(self.outfile):
            self.update_list = []
            self.is_updated_list = []
            for list_from_file in self._read_from_file(self.outfile):
                if isinstance(list_from_file, str):
                    self.update_list.append(list_from_file)
                    self.is_updated_list.append(False)
                else:
                    self.update_list.append(list_from_file[0])
                    self.is_updated_list.append(list_from_file[1])

        else:
            if self.parallel > 1:
                procs = []
                procs.append(threading.Thread(target=self.get_db_list))
                procs.append(threading.Thread(target=self.get_nomad_list))
                [p.start() for p in procs]
                [p.join() for p in procs]
            else:
                self.get_db_list()
                self.get_nomad_list()
            self.compare_lists()

    def update(self):
        self.prep_list()
        if self.do_download:
            self.download()


@client.command(
    help='Synchronizes the NOMAD database with the given external database.')
@click.argument('db_name', nargs=1, required=True)
@click.argument('root_url', nargs=1, required=True)
@click.option(
    '--outfile', default=None,
    help='File to read/write files missing in NOMAD database')
@click.option(
    '--nomadfile', default=None,
    help='File to read/write files in NOMAD database')
@click.option(
    '--dbfile', default=None,
    help='File to read/write files in given database')
@click.option(
    '--parallel', default=2,
    help='Number of processes to spawn to download/upload files')
@click.option(
    '--do-download', is_flag=True, default=False,
    help='Flag to automatically download downloaded files')
@click.option(
    '--do-upload', is_flag=True, default=False,
    help='Flag to automatically upload downloaded files')
@click.option(
    '--do-publish', is_flag=True, default=False,
    help='Flag to automatically publish upload')
@click.option(
    '--cleanup', is_flag=True, default=False,
    help='Flag to clean up downloaded files')
def synchdb(**kwargs):
    db = DbUpdater(**kwargs)
    db.update()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
The NOMAD meta-info allows to define schemas for physics data independent of the used
storage format. It allows to define physics quantities with types, complex shapes
(vetors, matrices, etc.), units, links, and descriptions. It allows to organize large
amounts of these quantities in containment hierarchies of extendable sections, references
between sections, and additional quantity categories.

NOMAD uses the meta-info to define all archive data, repository meta-data, (and encyclopedia
data). The meta-info provides a convenient Python interface to create,
manipulate, and access data. We also use it to map data to various storage formats,
including JSON, (HDF5), mongodb, and elastic search.

Starting example
----------------

.. code-block:: python

    from nomad.metainfo import MSection, Quantity, SubSection, Units

    class System(MSection):
        \'\'\'
        A system section includes all quantities that describe a single a simulated
        system (a.k.a. geometry).
        \'\'\'

        n_atoms = Quantity(
            type=int, description=\'\'\'
            A Defines the number of atoms in the system.
            \'\'\')

        atom_labels = Quantity(type=MEnum(ase.data.chemical_symbols), shape['n_atoms'])
        atom_positions = Quantity(type=float, shape=['n_atoms', 3], unit=Units.m)
        simulation_cell = Quantity(type=float, shape=[3, 3], unit=Units.m)
        pbc = Quantity(type=bool, shape=[3])

    class Run(MSection):
        section_system = SubSection(sub_section=System, repeats=True)


We define simple metainfo schema with two `sections` called ``System`` and ``Run``. Sections
allow to organize related data into, well, `sections`. Each section can have two types of
properties: `quantities` and `sub-sections`. Sections and their properties are defined with
Python classes and their attributes.

Each `quantity` defines a piece of data. Basic quantity attributes are its `type`, `shape`,
`unit`, and `description`.

`Sub-sections` allow to place section into each other and therefore allow to form containment
hierarchies or sections and the respective data in them. Basic sub-section attributes are
`sub_section`(i.e. a reference to the section definition of the sub-section) and `repeats`
(determines if a sub-section can be contained once or multiple times).

The above simply defines a schema, to use the schema and create actual data, we have to
instantiate the above classes:

.. code-block:: python

    run = Run()
    system = run.m_create(System)
    system.n_atoms = 3
    system.atom_labels = ['H', 'H', 'O']

    print(system.atom_labels)
    print(n_atoms = 3)


Section `instances` can be used like regular Python objects: quantities and sub-sections
can be set and access like any other Python attribute. Special meta-info methods, starting
with ``m_`` allow us to realize more complex semantics. For example ``m_create`` will
instantiate a sub-section and add it to the `parent` section in one step.

Another example for an ``m_``-method is:

.. code-block:: python

    run.m_to_json(indent=2)

This will serialize the data into JSON:

.. code-block:: JSON

    {
        "m_def" = "Run",
        "systems": [
            {
                "n_atoms" = 3,
                "atom_labels" = [
                    "H",
                    "H",
                    "O"
                ]
            }
        ]
    }

Definitions and Instances
-------------------------

As you already saw in the example, we first need to define what data can look like (schema),
before we can actually program with data. Because schema and data are often discussed in the
same context, it is paramount to clearly distingish between both. For example, if we
just say "system", it is unclear what we refer to. We could mean the idea of a system, i.e. all
possible systems, a data structure that comprises a lattice, atoms with their elements and
positions in the lattice. Or we mean a specific system of a specific calculation, with
a concrete set of atoms, real numbers for lattice vectors and atoms positions as concrete data.

The NOMAD Metainfo is just a collection of definition that describe what materials science
data could be (a schema). The NOMAD Archive is all the data that we extract from all data
provided to NOMAD. The data in the NOMAD Archive follows the definitions of the NOMAD metainfo.

Similarely, we need to distingish between the NOMAD Metainfo as a collection of definitions
and the Metainfo system that defines how to define a section or a quantity. In this sense,
we have a three layout model, were the Archive (data) is an instance of the Metainfo (schema)
and the Metainfo is an instance of the Metainfo system (schema of the schema).

This documentation describes the Metainfo by explaining the means of how to write down definitions
in Python. Conceptually we map the Metainfo system to Python language constructs, e.g.
a section definition is a Python class, a quantity a Python property, etc. If you are
familiar with databases, this is similar to what an object relational mapping (ORM) would do.


Common attributes of Metainfo Definitions
-----------------------------------------

In the example, you already saw the basic Python interface to the Metainfo. *Sections* are
represented in Python as objects. To define a section, you write a Python classes that inherits
from ``MSection``. To define sub-sections and quantities you use Python properties. The
definitions themselves are also objects derived from classes. For sub-sections and
quantities, you directly instantiate :class:`SubSection` and :class`Quantity`. For sections
there is a generated object derived from :class:`Section` that is available via
`m_def` from each `section class` and `section instance`.

These Python classes that are used to represent metainfo definitions form an inheritance
hierarchy to share common properties

.. autoclass:: Definition


Quantities
----------

Quantity definitions are the main building block of meta-info schemas. Each quantity
represents a single piece of data.

.. autoclass:: Quantity


Sections and Sub-Sections
-------------------------

.. _metainfo-sections:


The NOMAD Metainfo allows to create hierarchical (meta-)data structures. A hierarchy
is basically a tree, where *sections* make up the root and inner nodes of the tree,
and *quantities* are the leaves of the tree. In this sense a section can hold further
sub-sections (more branches of the tree) and quantities (leaves). We say a section can
have two types of *properties*: *sub-sections* and *quantities*.

There is a clear distinction between *section* and *sub-section*. The term *section*
refers to an object that holds data (*properties*) and the term *sub-section* refers to a
relation between *sections*, i.e. between the containing section and the contained
sections of the same type. Furthermore, we have to distinguish between *sections* and
*section definitions*, as well as *sub-sections* and *sub-section definitions*. A *section
definition* defines the possible properties of its instances, and an instantiating *section*
can store quantities and sub-sections according to its definition. A *sub-section*
definition defines a principle containment relationship between two *section definitions*
and a *sub-section* is a set of *sections* (contents) contained in another *section* (container).
*Sub-section definitions* are parts (*properties*) of the definition of containing sections.

.. autoclass:: Section

.. autoclass:: SubSection

.. _metainfo-categories:


References and Proxies
----------------------


Beside creating hierarchies (e.g. tree structures) with ``SubSection``, the metainfo
also allows to create cross references between sections and other sections or quantity
values:

.. code-block:: python

    class Calculation(MSection):
        system = Quantity(type=System.m_def)
        atom_labels = Quantity(type=System.atom_labels)

    calc = Calculation()
    calc.system = run.systems[-1]
    calc.atom_labels = run.systems[-1]


To define a reference, you define a normal quantity and simply use the section or quantity
that you want to reference as type. Then you can assign respective section instances
as values.

When in Python memory, quantity values that reference other sections simply contain a
Python reference to the respective `section instance`. However, upon serializing/storing
metainfo data, these references have to be represented differently.

Value references are a little different. When you read a value references it behaves like
the references value. Internally, we do not store the values, but a reference to the
section that holds the referenced quantity. Therefore, when you want to
assign a value reference, you use the section with the quantity and not the value itself.

Currently this metainfo implementation only supports references within a single
section hierarchy (e.g. the same JSON file). References are stored as paths from the
root section, over sub-sections, to the referenced section or quantity value. Each path segment is
the name of the sub-section or an index in a repeatable sub-section:
``/system/0`` or ``/system/0/atom_labels``.

References are automatically serialized by :py:meth:`MSection.m_to_dict`. When de-serializing
data with :py:meth:`MSection.m_from_dict` these references are not resolved right away,
because the references section might not yet be available. Instead references are stored
as :class:`MProxy` instances. These objects are automatically replaced by the referenced
object when a respective quantity is accessed.

.. autoclass:: MProxy

If you want to defined references, it might not be possible to define the referenced
section or quantity before hand, due to how Python definitions and imports work. In these
cases, you can use a proxy to reference the reference type:

.. code-block:: python

    class Calculation(MSection):
        system = Quantity(type=MProxy('System')
        atom_labels = Quantity(type=MProxy('System/atom_labels')

The strings given to ``MProxy`` are paths within the available definitions. The above example
works, if ``System`` and ``System/atom_labels`` are eventually defined in the same package.


Categories
----------

In the old meta-info this was known as `abstract types`.

Categories are defined with Python classes that have :class:`MCategory` as base class.
Their name and description is taken from the class's name and docstring. An example
category looks like this:

.. code-block:: python

    class CategoryName(MCategory):
        \'\'\' Category description \'\'\'
        m_def = Category(links=['http://further.explanation.eu'], categories=[ParentCategory])

Packages
--------

.. autoclass:: Package

.. _metainfo-custom-types:


Environments
------------

.. autoclass:: Environment


Custom data types
-----------------

.. autoclass:: DataType
    :members:

.. autoclass:: MEnum


.. _metainfo-reflection:

Reflection and custom data storage
----------------------------------

When manipulating metainfo data in Python, all data is represented as Python objects, where
objects correspond to `section instance` and their attributes to `quantity values` or
`section instances` of sub-sections. By defining sections with `section classes` each
of these Python objects already has an interface that allows to get/set quantities and
sub-sections. But often this interface is too limited, or the specific section and
quantity definitions are unknown when writing code.

.. autoclass:: MSection
    :members:

.. autoclass:: MetainfoError
.. autoclass:: DeriveError
.. autoclass:: MetainfoReferenceError

.. _metainfo-urls:

Resources
---------

.. autoclass:: MResource

A more complex example
----------------------

.. literalinclude:: ../nomad/metainfo/example.py
    :language: python

'''


from .metainfo import (
    MSectionBound,
    MSection,
    MCategory,
    Definition,
    Property,
    Quantity,
    SubSection,
    Section,
    Category,
    Package,
    Environment,
    MEnum,
    Datetime,
    Capitalized,
    MProxy,
    MetainfoError,
    DeriveError,
    MetainfoReferenceError,
    DataType,
    Reference,
    QuantityReference,
    Datetime,
    Unit,
    JSON,
    Dimension,
    MResource,
    m_package,
    Annotation,
    DefinitionAnnotation,
    SectionAnnotation,
    SectionProxy,
    derived,
    constraint,
    units)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
The *parsing* module is an interface for the existing NOMAD-coe parsers.
This module redefines some of the old NOMAD-coe python-common functionality to create a
more coherent interface to the parsers.

Assumption about parsers
------------------------
For now, we make a few assumption about parsers
- they always work on the same *meta-info* version
- they have no conflicting python requirements
- they can be loaded at the same time and can be used within the same python process
- they are uniquely identified by a GIT URL and publicly accessible
- their version is uniquely identified by a GIT commit SHA

Each parser is defined via an instance of :class:`Parser`. The implementation :class:`LegacyParser` is used for most NOMAD-coe parsers.

.. autoclass:: nomad.parsing.Parser
    :members:

The are sub-classes for parsers with special purposes.

.. autoclass:: nomad.parsing.Parser
.. autoclass:: nomad.parsing.MatchingParser
.. autoclass:: nomad.parsing.MissingParser
.. autoclass:: nomad.parsing.BrokenParser
.. autoclass:: nomad.parsing.TemplateParser
.. autoclass:: nomad.parsing.GenerateRandomParser
.. autoclass:: nomad.parsing.ChaosParser
.. autoclass:: nomad.parsing.EmptyParser


The implementation :class:`LegacyParser` is used for most NOMAD-coe parsers.

.. autoclass:: nomad.parsing.LegacyParser


The parser definitions are available via the following two variables.

.. autodata:: nomad.parsing.parsers.parsers
.. autodata:: nomad.parsing.parsers.parser_dict

Parsers are reused for multiple calculations.

Parsers and calculation files are matched via regular expressions.

.. autofunction:: nomad.parsing.parsers.match_parser

Parsers in NOMAD-coe use a *backend* to create output. There are different NOMAD-coe
basends. In nomad@FAIRDI, we only currently only use a single backed. The following
classes provide a interface definition for *backends* as an ABC and a concrete implementation
based on nomad@fairdi's metainfo:

.. autoclass:: nomad.parsing.Backend
    :members:
'''

from nomad.parsing.parser import Parser, BrokenParser, MissingParser, MatchingParser, FairdiParser
from nomad.parsing.artificial import TemplateParser, GenerateRandomParser, ChaosParser, EmptyParser



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import numpy as np
import logging
import xarray
import json
import re

from nomad.units import ureg
from nomad.parsing.parser import FairdiParser
from nomad.parsing.file_parser import FileParser

from .metainfo import m_env
from nomad.datamodel.metainfo.common_dft import Run, Method, System, Workflow,\
    SingleConfigurationCalculation, EnergyContribution, StressTensorContribution,\
    AtomStressContribution, XCFunctionals
from fhivibesparser.metainfo.fhi_vibes import x_fhi_vibes_section_attributes,\
    x_fhi_vibes_section_metadata, x_fhi_vibes_section_atoms, x_fhi_vibes_section_MD,\
    x_fhi_vibes_section_calculator, x_fhi_vibes_section_calculator_parameters,\
    x_fhi_vibes_section_vibes, x_fhi_vibes_section_relaxation,\
    x_fhi_vibes_section_relaxation_kwargs, x_fhi_vibes_section_settings,\
    x_fhi_vibes_section_phonopy


class XarrayParser(FileParser):
    def __init__(self):
        super().__init__()
        self.re_index = re.compile(r'(.+?)\[(\d+)\]')
        self._raw_metadata = dict()

    @property
    def dataset(self):
        if self._file_handler is None:
            try:
                self._file_handler = xarray.open_dataset(self.mainfile)
                self._raw_metadata = json.loads(self._file_handler.attrs.get('raw_metadata', '{}'))
            except Exception:
                self.logger.error('Error reading trajectory file.')
                pass

        return self._file_handler

    def parse(self, key):
        key = key.strip('/')
        val = self.dataset
        for section in key.split('/'):
            indexed = re.match(self.re_index, section)
            if indexed:
                section = indexed.group(1)
                index = indexed.group(2)
            if section == 'raw_metadata':
                val = self._raw_metadata
            elif section == 'attrs':
                val = self.dataset.attrs
            else:
                val = val.get(section)
            try:
                val = val[int(index)] if indexed else val
            except Exception:
                val = None
            if val is None:
                break
        if isinstance(val, xarray.DataArray):
            val = np.array(val, dtype=val.dtype)

        self._results[key] = val


class FHIVibesParser(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/fhi-vibes', code_name='FHI-vibes',
            code_homepage='https://vibes.fhi-berlin.mpg.de/',
            mainfile_name_re=(r'^.*\.(nc)$'), mainfile_mime_re=r'(application/x-hdf)',
            mainfile_binary_header_re=br'^\x89HDF')

        self.parser = XarrayParser()
        self._metainfo_env = m_env

        self._units = {
            'volume': ureg.angstrom ** 3, 'displacements': ureg.angstrom, 'velocities': ureg.angstrom / ureg.fs,
            'momenta': ureg.eV * ureg.fs / ureg.angstrom, 'force_constants': ureg.eV / ureg.angstrom ** 2,
            'forces_harmonic': ureg.eV / ureg.angstrom, 'forces': ureg.eV / ureg.angstrom, 'stress': ureg.eV / ureg.angstrom ** 3,
            'energy_potential_harmonic': ureg.eV, 'sigma_per_sample': None,
            'pressure': ureg.eV / ureg.angstrom ** 3, 'temperature': ureg.K, 'pressure_kinetic': ureg.eV / ureg.angstrom ** 3,
            'pressure_potential': ureg.eV / ureg.angstrom ** 3, 'aims_uuid': None, 'energy': ureg.eV,
            'heat_flux': ureg.amu / ureg.fs ** 3, 'heat_flux_harmonic': ureg.amu / ureg.fs ** 3,
            'heat_flux_0_harmonic': ureg.amu / ureg.fs ** 3, 'mass': ureg.amu, 'length': ureg.angstrom, 'time': ureg.fs}

    @property
    def n_frames(self):
        if self.calculation_type == 'phonon':
            return 1
        return len(self.parser.get('positions'))

    def parse_configurations(self):

        def parse_system(n_frame):
            sec_system = sec_run.m_create(System)
            sec_system.atom_labels = self.parser.get('attrs').get('symbols')
            sec_system.atom_positions = self.parser.get('positions', unit=self._units.get('length'))[n_frame]
            sec_system.lattice_vectors = self.parser.get('cell', unit=self._units.get('length'))[n_frame]
            sec_system.configuration_periodic_dimensions = self.parser.get('attrs/raw_metadata/atoms/pbc')
            velocities = self.parser.get('velocities', unit=self._units.get('velocities'))
            if velocities is not None:
                sec_system.atom_velocities = velocities[n_frame]
            return sec_system

        def parse_scc(n_frame):
            sec_scc = sec_run.m_create(SingleConfigurationCalculation)

            if self.calculation_type == 'molecular_dynamics':
                sec_scc.time_step = n_frame
                # TODO metainfo should be in common
                sec_scc.x_fhi_vibes_MD_time = n_frame * timestep

            for key in ['kinetic', 'potential']:
                val = self.parser.get('energy_%s' % key, unit=self._units.get('energy'))
                if val is not None:
                    sec_energy = sec_scc.m_create(EnergyContribution)
                    sec_energy.energy_contribution_kind = key
                    sec_energy.energy_contribution_value = val[n_frame]

                val = self.parser.get('stress_%s' % key, unit=self._units.get('stress'))
                if val is not None:
                    sec_stress = sec_scc.m_create(StressTensorContribution)
                    sec_stress.stress_tensor_contribution_kind = key
                    sec_stress.stress_tensor_contribution_value = val[n_frame]

                val = self.parser.get('stresses_%s' % key, unit=self._units.get('stress'))
                if val is not None:
                    sec_atom_stress = sec_scc.m_create(AtomStressContribution)
                    sec_atom_stress.atom_stress_contribution_kind = key
                    sec_atom_stress.atom_stress_contribution_value = val[n_frame]

            calculation_quantities = [
                'volume', 'displacements', 'momenta', 'forces_harmonic',
                'forces', 'stress', 'energy_potential_harmonic', 'sigma_per_sample',
                'pressure', 'temperature', 'pressure_kinetic', 'pressure_potential', 'aims_uuid',
                'heat_flux', 'heat_flux_harmonic', 'heat_flux_0_harmonic']
            for key in calculation_quantities:
                val = self.parser.get(key, unit=self._units.get(key, None))
                if val is None:
                    continue

                # TODO figure out what shape of output fc
                if key.startswith('force_constants'):
                    continue
                if key.startswith('forces'):
                    key = 'atom_%s' % key
                elif key == 'stress':
                    key = 'stress_tensor'

                if key in ['temperature', 'pressure', 'atom_forces', 'stress_tensor']:
                    setattr(sec_scc, key, val[n_frame])
                else:
                    setattr(sec_scc, 'x_fhi_vibes_%s' % key, val[n_frame])

            return sec_scc

        sec_atrr = self.archive.section_run[-1].section_method[-1].x_fhi_vibes_section_attributes[-1]
        timestep = sec_atrr.x_fhi_vibes_attributes_timestep
        for n_frame in range(self.n_frames):
            if self.calculation_type == 'single_point':
                sec_run = self.archive.section_run[n_frame]
                # we can only do this for single point where we have separate section_runs
                # for each frame
                sec_run.raw_id = self.parser.get('aims_uuid')[n_frame]
            else:
                # TODO aims_uuid is in x_fhi_vibes_aims_uuid, this should be changed
                sec_run = self.archive.section_run[-1]

            sec_system = parse_system(n_frame)
            sec_scc = parse_scc(n_frame)
            sec_scc.single_configuration_calculation_to_system_ref = sec_system
            sec_scc.single_configuration_to_calculation_method_ref = sec_run.section_method[-1]

        # force constants
        for key in ['force_constants', 'force_constants_remapped']:
            val = self.parser.get(key, unit=self._units.get('force_constants'))
            if val is not None:
                setattr(sec_scc, 'x_fhi_vibes_%s' % key, val)

    def parse_method(self, n_run):
        def parse_xc_functional():
            # TODO This is a temporary fix to circumvent the normalization tests failure
            # due to missing xc functional information but this should be fetched directly
            # from the underlying calculation
            xc_type_map = {'PW': 'C'}
            calculator_parameters = self.parser.get(
                'attrs/raw_metadata/calculator/calculator_parameters', {})
            xc_functional = calculator_parameters.get('xc', '').upper()
            xc_functional_info = re.match(r'(\w+)\S+?((?:LDA|GGA|MGGA|HYB_GGA|HYB_MGGA))', xc_functional)
            if xc_functional_info:
                xc_name = xc_functional_info.group(1)
                xc_type = xc_type_map.get(xc_name, None)
                if xc_type is None:
                    self.logger.error('Cannot resolve XC functional.')
                    return
                sec_xc_functional = sec_method.m_create(XCFunctionals)
                sec_xc_functional.XC_functional_name = '%s_%s_%s' % (
                    xc_functional_info.group(2), xc_type, xc_functional_info.group(1))

        def parse_atoms(section, atoms):
            for key, val in atoms.items():
                # why is the formatting of symbols and masses different for atoms?
                if key in ['symbols', 'masses']:
                    if isinstance(val[0], list):
                        val_flattened = []
                        for val_i in val:
                            val_flattened.extend([val_i[1]] * val_i[0])
                        val = val_flattened
                    if key == 'masses':
                        val = val * self._units.get('mass')
                elif key in ['positions', 'cell']:
                    val = val * self._units.get('length')
                elif key == 'velocities':
                    val = val * self._units.get('length') / self._units.get('time')
                setattr(section, 'x_fhi_vibes_atoms_%s' % key, val)

        def parse_metadata():
            metadata = self.parser.get('attrs/raw_metadata')
            sec_metadata = sec_attrs.m_create(x_fhi_vibes_section_metadata)
            for key, val in metadata.items():
                if key == 'MD':
                    sec_md = sec_metadata.m_create(x_fhi_vibes_section_MD)
                    for md_key in val.keys():
                        setattr(
                            sec_md, 'x_fhi_vibes_MD_%s' % md_key.replace('-', '_'),
                            val[md_key])
                elif key == 'relaxation':
                    sec_relaxation = sec_metadata.m_create(x_fhi_vibes_section_relaxation)
                    for relaxation_key in val.keys():
                        if relaxation_key == 'kwargs':
                            sec_kwargs = sec_relaxation.m_create(x_fhi_vibes_section_relaxation_kwargs)
                            for kwargs_key in val['kwargs']:
                                setattr(
                                    sec_kwargs, 'x_fhi_vibes_relaxation_kwargs_%s' % kwargs_key,
                                    val['kwargs'][kwargs_key])
                        else:
                            setattr(
                                sec_relaxation, 'x_fhi_vibes_relaxation_%s' % relaxation_key.replace('-', '_'),
                                val[relaxation_key])
                elif key == 'Phonopy':
                    sec_phonopy = sec_metadata.m_create(x_fhi_vibes_section_phonopy)
                    for phonopy_key in val.keys():
                        if phonopy_key == 'primitive':
                            sec_primitive = sec_phonopy.m_create(x_fhi_vibes_section_atoms)
                            parse_atoms(sec_primitive, val['primitive'])
                        else:
                            setattr(sec_phonopy, 'x_fhi_vibes_phonopy_%s' % phonopy_key, val[phonopy_key])
                elif key == 'calculator':
                    sec_calculator = sec_metadata.m_create(x_fhi_vibes_section_calculator)
                    sec_calculator.x_fhi_vibes_calculator = metadata['calculator']['calculator']
                    sec_calculator_parameters = sec_calculator.m_create(x_fhi_vibes_section_calculator_parameters)
                    for calc_key in val['calculator_parameters'].keys():
                        if calc_key == 'use_pimd_wrapper':
                            val['calculator_parameters'][calc_key] = str(val['calculator_parameters'][calc_key])
                        setattr(
                            sec_calculator_parameters, 'x_fhi_vibes_calculator_parameters_%s' % calc_key,
                            val['calculator_parameters'][calc_key])
                elif key in ['atoms', 'primitive', 'supercell']:
                    sec_atoms = sec_metadata.m_create(x_fhi_vibes_section_atoms)
                    sec_atoms.x_fhi_vibes_atoms_kind = key
                    parse_atoms(sec_atoms, val)
                elif key == 'vibes':
                    sec_vibes = sec_metadata.m_create(x_fhi_vibes_section_vibes)
                    for vibes_key in val.keys():
                        setattr(sec_vibes, 'x_fhi_vibes_%s' % vibes_key, val[vibes_key])
                elif key == 'settings':
                    sec_settings = sec_metadata.m_create(x_fhi_vibes_section_settings)
                    for settings_key in val.keys():
                        setattr(sec_settings, 'x_fhi_vibes_settings_%s' % settings_key, val[settings_key])
                else:
                    setattr(sec_metadata, key, val)

        sec_method = self.archive.section_run[n_run].m_create(Method)

        parse_xc_functional()

        sec_attrs = sec_method.m_create(x_fhi_vibes_section_attributes)

        time_units = {'ns': ureg.ns, 'fs': ureg.fs, 'ps': ureg.ps}
        attrs = self.parser.get('attrs')
        for key, val in attrs.items():
            if key == 'raw_metadata':
                parse_metadata()
            elif key.startswith('atoms_'):
                sec_atoms = sec_attrs.m_create(x_fhi_vibes_section_atoms)
                sec_atoms.x_fhi_vibes_atoms_kind = key
                atoms = json.loads(val)
                sec_atoms.x_fhi_vibes_atoms_natoms = len(atoms['positions'])
                parse_atoms(sec_atoms, atoms)
                setattr(sec_attrs, 'x_fhi_vibes_attributes_number_of_%s' % key, len(atoms['positions']))
            else:
                if key == 'masses':
                    val = val * self._units.get('mass')
                elif key == 'timestep':
                    val = val * time_units.get(attrs.get('time_unit').lower(), self._units.get('time'))
                setattr(sec_attrs, 'x_fhi_vibes_attributes_%s' % key, val)

        # we need this information for force constants
        n_atoms_supercell = sec_attrs.x_fhi_vibes_attributes_number_of_atoms_supercell
        if n_atoms_supercell:
            sec_attrs.x_fhi_vibes_attributes_force_constants_remapped_size = n_atoms_supercell * 3

    def init_parser(self):
        self.parser.mainfile = self.filepath
        self.parser.logger = self.logger

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.logger = logger if logger is not None else logging.getLogger(__name__)
        self.maindir = os.path.dirname(self.filepath)

        self.init_parser()

        metadata = self.parser.get('attrs/raw_metadata')

        if 'MD' in metadata:
            self.calculation_type = 'molecular_dynamics'
        elif 'relaxation' in metadata:
            self.calculation_type = 'geometry_optimization'
        elif 'Phonopy' in metadata:
            self.calculation_type = 'phonon'
        else:
            # the single point workflow in vibes means multiple separate calculations on
            # on same material (stoichiometry, number of atoms) but may differ on
            # on structure (lattice, positions). This means we need to create separate
            # section runs
            self.calculation_type = 'single_point'

        if self.calculation_type == 'single_point':
            for _ in range(self.n_frames):
                self.archive.m_create(Run)
        else:
            self.archive.m_create(Run)

        for n_run, sec_run in enumerate(self.archive.section_run):
            sec_run.program_name = 'FHI-vibes'
            sec_run.program_version = metadata['vibes']['version']

            if metadata['calculator']['calculator'].lower() == 'aims':
                sec_run.program_basis_set_type = 'numeric AOs'

            self.parse_method(n_run)

        # TODO For single_point, we can only have workflow for one vibes single point frame
        # as workflow is not repeating in metainfo.
        # To resolve this, we can redefine single_point workflow to be consistent with
        # the idea of vibes single point but I do not like it.
        sec_workflow = self.archive.m_create(Workflow)
        sec_workflow.workflow_type = self.calculation_type
        sec_workflow.calculator = metadata['calculator']['calculator']

        self.parse_configurations()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public

m_package = Package(
    name='turbomole_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='turbomole.nomadmetainfo.json'))


class x_turbomole_section_eigenvalues_GW(MSection):
    '''
    section for the eigenvalues of a GW calculation (at present only pertubative G0W0)
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_turbomole_section_eigenvalues_GW'))

    x_turbomole_eigenvalue_correlation_perturbativeGW = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Correlation energy at a given eigenstate from perturbative GW
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_eigenvalue_correlation_perturbativeGW'))

    x_turbomole_eigenvalue_ExactExchange_perturbativeGW = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Exact exchange energy at given eigenstate from perturbative GW
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_eigenvalue_ExactExchange_perturbativeGW'))

    x_turbomole_eigenvalue_ExchangeCorrelation_perturbativeGW = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Self-energy at a given eigenstate from perturbative GW
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_eigenvalue_ExchangeCorrelation_perturbativeGW'))

    x_turbomole_eigenvalue_ks_ExchangeCorrelation = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        KS exchange correlation energy at a given eigenstate needed to calculate the
        quasi-particle energy in perturbative GW
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_eigenvalue_ks_ExchangeCorrelation'))

    x_turbomole_eigenvalue_ks_GroundState = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        KS ground state energy at a given eigenstate needed in perturbative GW
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_eigenvalue_ks_GroundState'))

    x_turbomole_eigenvalue_quasiParticle_energy = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Quasiparticle energy at a given eigenstate from perturbative GW
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_eigenvalue_quasiParticle_energy'))

    x_turbomole_ExchangeCorrelation_perturbativeGW_derivation = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        TODO:
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_ExchangeCorrelation_perturbativeGW_derivation'))

    x_turbomole_Z_factor = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        TODO:
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_Z_factor'))


class x_turbomole_section_functionals(MSection):
    '''
    section for one list of XC functionals
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_turbomole_section_functionals'))

    x_turbomole_controlInOut_grid_integration_cells = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The integration cells
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlInOut_grid_integration_cells'))

    x_turbomole_controlInOut_grid_integration = Quantity(
        type=str,
        shape=[],
        description='''
        type of the used grid integration
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlInOut_grid_integration'))

    x_turbomole_controlInOut_grid_partition_func = Quantity(
        type=str,
        shape=[],
        description='''
        Type of the partition function used
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlInOut_grid_partition_func'))

    x_turbomole_controlInOut_grid_partition_sharpness = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Sharpness of the partition function
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlInOut_grid_partition_sharpness'))

    x_turbomole_controlInOut_grid_points_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Grid points number
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlInOut_grid_points_number'))

    x_turbomole_controlInOut_grid_radial_grid_size = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The size of the radial grid
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlInOut_grid_radial_grid_size'))

    x_turbomole_controlInOut_grid_radial_integration = Quantity(
        type=str,
        shape=[],
        description='''
        The radial integration type
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlInOut_grid_radial_integration'))

    x_turbomole_controlInOut_grid_size = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        The size of the used grid
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlInOut_grid_size'))

    x_turbomole_XC_functional_type = Quantity(
        type=str,
        shape=[],
        description='''
        XC functional type
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_XC_functional_type'))


class section_method(public.section_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_method'))

    x_turbomole_controlIn_atom_label = Quantity(
        type=str,
        shape=[],
        description='''
        The label of the atoms in the system
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_atom_label'))

    x_turbomole_controlIn_atom_number = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        The number of atoms in the system
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_atom_number'))

    x_turbomole_controlIn_basis_status = Quantity(
        type=str,
        shape=[],
        description='''
        Status mean here ON or OFF
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_basis_status'))

    x_turbomole_controlIn_cartesian_status = Quantity(
        type=str,
        shape=[],
        description='''
        Status mean here ON or OFF
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_cartesian_status'))

    x_turbomole_controlIn_damping_parameter_min = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_damping_parameter_min'))

    x_turbomole_controlIn_damping_parameter_start = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_damping_parameter_start'))

    x_turbomole_controlIn_damping_parameter_step = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_damping_parameter_step'))

    x_turbomole_controlIn_dipole_status = Quantity(
        type=str,
        shape=[],
        description='''
        Status mean here ON or OFF
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_dipole_status'))

    x_turbomole_controlIn_global_status = Quantity(
        type=str,
        shape=[],
        description='''
        Status mean here ON or OFF
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_global_status'))

    x_turbomole_controlIn_hessian_status = Quantity(
        type=str,
        shape=[],
        description='''
        Status mean here ON or OFF
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_hessian_status'))

    x_turbomole_controlIn_interconversion_status = Quantity(
        type=str,
        shape=[],
        description='''
        Status mean here ON or OFF
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_interconversion_status'))

    x_turbomole_controlIn_number_of_integral_stored = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_number_of_integral_stored'))

    x_turbomole_controlIn_operating_system = Quantity(
        type=str,
        shape=[],
        description='''
        The kind of operating system
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_operating_system'))

    x_turbomole_controlIn_pople_kind = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_pople_kind'))

    x_turbomole_controlIn_scf_conv = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_scf_conv'))

    x_turbomole_controlIn_scf_iter_limit = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_scf_iter_limit'))

    x_turbomole_controlIn_scfintunit_file = Quantity(
        type=str,
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_scfintunit_file'))

    x_turbomole_controlIn_scfintunit_size = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_scfintunit_size'))

    x_turbomole_controlIn_scfintunit_unit = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_scfintunit_unit'))

    x_turbomole_controlIn_symmetry = Quantity(
        type=str,
        shape=[],
        description='''
        The given symmetry of the system
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_symmetry'))

    x_turbomole_controlIn_time_for_integral_calc = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        -
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_controlIn_time_for_integral_calc'))

    x_turbomole_dft_d3_version = Quantity(
        type=str,
        shape=[],
        description='''
        version of the DFT-D3 van-der-Waals correction that is used
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_dft_d3_version'))

    x_turbomole_functional_type_correlation = Quantity(
        type=str,
        shape=[],
        description='''
        type of the used correlation functional
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_functional_type_correlation'))

    x_turbomole_functional_type_exchange = Quantity(
        type=str,
        shape=[],
        description='''
        type of the used exchange functional
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_functional_type_exchange'))

    x_turbomole_geometry_optimization_cycle_index = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        By default Turbomole only keeps the output of the final iteration once the
        geometry has been converged, thus the entire optimization trajectory cannot be
        rebuild in most cases. Instead, this value contains the optimization cycle index
        to indicate how many iterations have preceded this one.
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_geometry_optimization_cycle_index'))

    x_turbomole_gw_approximation = Quantity(
        type=str,
        shape=[],
        description='''
        The employed GW approximation.
        ''',
        categories=[public.settings_XC, public.settings_post_hartree_fock, public.settings_GW, public.settings_potential_energy_surface],
        a_legacy=LegacyDefinition(name='x_turbomole_gw_approximation'))

    x_turbomole_gw_eta_factor = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        [TO BE VERIFIED]Infinitesimal complex energy shift. Negative value switches to
        calculating at that value but extrapolating to 0 in linear approximation.
        ''',
        categories=[public.settings_XC, public.settings_post_hartree_fock, public.settings_GW, public.settings_potential_energy_surface],
        a_legacy=LegacyDefinition(name='x_turbomole_gw_eta_factor'))

    x_turbomole_gw_use_rpa_response = Quantity(
        type=bool,
        shape=[],
        description='''
        If true, the pure RPA response function is calculated. Otherwise, the TDDFT
        response function is calculated and used to screen the coulomb interaction.
        ''',
        categories=[public.settings_XC, public.settings_post_hartree_fock, public.settings_GW, public.settings_potential_energy_surface],
        a_legacy=LegacyDefinition(name='x_turbomole_gw_use_rpa_response'))

    x_turbomole_uhfmo_type = Quantity(
        type=str,
        shape=[],
        description='''
        Type of UHF molecular orbital
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_uhfmo_type'))

    x_turbomole_section_functionals = SubSection(
        sub_section=SectionProxy('x_turbomole_section_functionals'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_turbomole_section_functionals'))


class section_scf_iteration(public.section_scf_iteration):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_scf_iteration'))

    x_turbomole_damping_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Damping of the two-electron contributions to Fock matrix in the present SCF
        iteration
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_damping_scf_iteration'))

    x_turbomole_delta_eigenvalues = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        change of the eigenvalues in the current SCF iteration
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_turbomole_delta_eigenvalues'))

    x_turbomole_energy_1electron_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Total energy contribution from one-electron integrals
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_turbomole_energy_1electron_scf_iteration'))

    x_turbomole_energy_2electron_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Total energy contribution from two-electron integrals
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_turbomole_energy_2electron_scf_iteration'))

    x_turbomole_norm_diis_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Norm of the DIIS error in an SCF-iteration
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_norm_diis_scf_iteration'))

    x_turbomole_norm_fia_orbital_scf_iteration = Quantity(
        type=str,
        shape=[],
        description='''
        orbital with the largest residual norm for the Fia block in this iteration
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_norm_fia_orbital_scf_iteration'))

    x_turbomole_norm_fia_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Maximal resid. norm for Fia-block in an SCF-iteration
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_norm_fia_scf_iteration'))

    x_turbomole_norm_fock_orbital_scf_iteration = Quantity(
        type=str,
        shape=[],
        description='''
        orbital with the largest residual Fock norm in this iteration
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_norm_fock_orbital_scf_iteration'))

    x_turbomole_norm_fock_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Maximal resid. fock norm in an SCF-iteration
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_norm_fock_scf_iteration'))


class section_eigenvalues(public.section_eigenvalues):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_eigenvalues'))

    x_turbomole_eigenvalues_irreducible_representation = Quantity(
        type=np.dtype('U'),
        shape=['number_of_spin_channels', 'number_of_eigenvalues_kpoints', 'number_of_eigenvalues'],
        description='''
        Irreducible representation the eigenstates belong to.
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_eigenvalues_irreducible_representation'))

    x_turbomole_section_eigenvalues_GW = SubSection(
        sub_section=SectionProxy('x_turbomole_section_eigenvalues_GW'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_turbomole_section_eigenvalues_GW'))


class section_sampling_method(public.section_sampling_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_sampling_method'))

    x_turbomole_geometry_optimization_geometry_change_rms = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        geometry optimization convergence criterion - Root Mean Square of displacements
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_geometry_optimization_geometry_change_rms'))

    x_turbomole_geometry_optimization_threshold_force_rms = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='newton',
        description='''
        geometry optimization convergence criterion - Root Mean Square of forces
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_geometry_optimization_threshold_force_rms'))

    x_turbomole_geometry_optimization_trustregion_initial = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        geometry optimization trust region - initial radius
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_geometry_optimization_trustregion_initial'))

    x_turbomole_geometry_optimization_trustregion_max = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        geometry optimization trust region - maximum radius
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_geometry_optimization_trustregion_max'))

    x_turbomole_geometry_optimization_trustregion_min = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        geometry optimization trust region - minimum radius
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_geometry_optimization_trustregion_min'))


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_single_configuration_calculation'))

    x_turbomole_module = Quantity(
        type=str,
        shape=[],
        description='''
        The name of the Turbomole module used for this single configuration calculation.
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_module'))

    x_turbomole_potential_energy_final = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Final potential energy
        ''',
        categories=[public.energy_value, public.energy_component],
        a_legacy=LegacyDefinition(name='x_turbomole_potential_energy_final'))

    x_turbomole_vibrations_infrared_activity = Quantity(
        type=bool,
        shape=['x_turbomole_vibrations_num_modes'],
        description='''
        IR activicity for vibration modes
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_vibrations_infrared_activity'))

    x_turbomole_vibrations_raman_activity = Quantity(
        type=bool,
        shape=['x_turbomole_vibrations_num_modes'],
        description='''
        Raman activicity for vibration modes
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_vibrations_raman_activity'))

    x_turbomole_vibrations_intensities = Quantity(
        type=np.dtype(np.float64),
        shape=['x_turbomole_vibrations_num_modes'],
        description='''
        IR Intensity for each vibrational normal mode
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_vibrations_intensities'))

    x_turbomole_vibrations_mode_energies = Quantity(
        type=np.dtype(np.float64),
        shape=['x_turbomole_vibrations_num_modes'],
        description='''
        Excitation energy associated with the vibrational normal modes.
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_vibrations_mode_energies'))

    x_turbomole_vibrations_normal_modes = Quantity(
        type=np.dtype(np.float64),
        shape=['x_turbomole_vibrations_num_modes', 'number_of_atoms', 3],
        description='''
        Nuclear displacements for each vibrational normal mode
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_vibrations_normal_modes'))

    x_turbomole_vibrations_num_modes = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of vibrational normal modes
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_vibrations_num_modes'))

    x_turbomole_virial_theorem = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Final value from the virial theorem
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_virial_theorem'))

    x_turbomole_wave_func_norm = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Final Wave Function Norm
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_wave_func_norm'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_turbomole_nodename = Quantity(
        type=str,
        shape=[],
        description='''
        compute node
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_nodename'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    x_turbomole_pceem_charges = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms'],
        description='''
        Charges of the point charges in the unit cell used by the PCEEM embedding model
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_pceem_charges'))

    x_turbomole_pceem_max_multipole = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Maximum multipole moment used in the PCEEM embedding
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_pceem_max_multipole'))

    x_turbomole_pceem_min_separation_cells = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Minimum separation between cells in PCEEM embedding for periodic fast multipole
        treatment
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_pceem_min_separation_cells'))

    x_turbomole_pceem_multipole_precision = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Multipole precision parameter for PCEEM embedding
        ''',
        a_legacy=LegacyDefinition(name='x_turbomole_pceem_multipole_precision'))


m_package.__init_metainfo__()
